"""SC Controller - Daemon class."""

import json
import logging
import os
import pkgutil
import signal
import stat
import subprocess
import sys
import threading
import time
import traceback
from socketserver import StreamRequestHandler, ThreadingMixIn, UnixStreamServer

from scc import drivers
from scc.actions import Action
from scc.cemuhook_server import CemuhookServer
from scc.config import Config
from scc.constants import CPAD, DAEMON_VERSION, DPAD, LEFT, RIGHT, RSTICK, STICK, HapticPos, SCButtons
from scc.controller import HapticData
from scc.custom import load_custom_module
from scc.device_monitor import create_device_monitor
from scc.gestures import GestureDetector
from scc.lib import xinput
from scc.lib import xwrappers as X
from scc.lib.daemon import Daemon
from scc.mapper import Mapper
from scc.menu_data import MenuData
from scc.parser import TalkingActionParser
from scc.poller import Poller
from scc.profile import Profile
from scc.scheduler import Scheduler
from scc.tools import clamp, find_binary, find_menu, find_profile, nameof, set_logging_level, shjoin, shsplit
from scc.uinput import CannotCreateUInputException

log = logging.getLogger("SCCDaemon")
tlog = logging.getLogger("Socket Thread")


class ThreadingUnixStreamServer(ThreadingMixIn, UnixStreamServer):
	daemon_threads = True


class SCCDaemon(Daemon):
	def __init__(self, piddile, socket_file):
		set_logging_level(True, True)
		Daemon.__init__(self, piddile)
		Config()  # Generates ~/.config/scc and default config if needed
		self.started = False
		self.exiting = False
		self.socket_file = socket_file
		self.poller = Poller()
		self.dev_monitor = create_device_monitor(self)
		self.scheduler = Scheduler()
		self.xdisplay = None
		self.sserver = None  # UnixStreamServer instance
		self.errors = []
		self.alone = False  # Set by launching script from --alone flag
		self.custom_py_loaded = False
		self.osd_daemon = None
		self.default_profile = None
		self.autoswitch_daemon = None
		# TODO: Use osd_ids for all menus
		self.osd_ids = {}
		self.controllers = []
		self.mainloops = [self.poller.poll, self.scheduler.run]
		self.rescan_cbs = []
		self.on_exit_cbs = []
		self.subprocs = []
		self.lock = threading.Lock()
		self.cemuhook = None
		self.default_mapper = None
		self.free_mappers = []
		self.clients = set()
		self.cwd = os.getcwd()

	def init_drivers(self):
		"""Search and initialize all controller drivers.

		See __init__.py in scc.drivers.
		"""
		log.debug("Initializing drivers...")
		cfg = Config()
		self._to_start = set()  # del-eted later by start_drivers
		to_init = []
		for importer, modname, ispkg in pkgutil.walk_packages(path=drivers.__path__, onerror=lambda x: None):
			if not ispkg and modname != "driver":
				if modname == "usb" or cfg["drivers"].get(modname):
					# 'usb' driver has to be always active
					mod = getattr(__import__(f"scc.drivers.{modname}").drivers, modname)
					if hasattr(mod, "init"):
						to_init.append(mod)
				else:
					log.warning("Skipping disabled driver '%s'", modname)

		from scc.drivers import MOD_INIT_ORDER as order

		index_fn = lambda n: order.index(n) if n in order else 1024
		sort_fn = lambda m: index_fn(m.__name__)

		for mod in sorted(to_init, key=sort_fn):
			if mod.init(self, cfg):
				if hasattr(mod, "start"):
					self._to_start.add(mod.start)

	def init_default_mapper(self):
		"""default_mapper is persistent mapper assigned to first Controller instance.
		Even if all controllers are removed, this mapper stays active. This is
		needed so various stuff (mainlg GUI) doesn't need to check if there is
		any controller connected all the time.
		"""
		# But, despite all above, it's just mapper as every other :)
		return self.init_mapper()

	def set_default_profile(self, profile_file):
		"""Sets profile that is used for first available controller
		"""
		self.default_profile = profile_file

	def start_drivers(self):
		for s in self._to_start:
			s(self)
		del self._to_start

	def stop_drivers(self):
		for s in self.drivers_to_stop:
			s(self)

	def get_poller(self):
		"""Returns poller that can be used for polling file descriptors"""
		return self.poller

	def get_device_monitor(self):
		"""Returns device monitor that can be used to listen for device adding and removals
		"""
		return self.dev_monitor

	def get_scheduler(self):
		"""Returns scheduler instance"""
		return self.scheduler

	def add_mainloop(self, fn):
		"""Adds function that is called in every mainloop iteration.
		Can be called only durring initialization, in driver 'init' method.
		"""
		if fn not in self.mainloops:
			self.mainloops.append(fn)

	def remove_mainloop(self, fn):
		"""Removes function added by add_mainloop
		"""
		if fn in self.mainloops:
			self.mainloops.remove(fn)

	def add_on_exit(self, fn):
		"""Adds function that is called just before daemon is stopped.
		Usefull for cleanup.
		"""
		if fn not in self.on_exit_cbs:
			self.on_exit_cbs.append(fn)

	def add_on_rescan(self, fn):
		"""Adds function that is called when `Rescan.` message is received.
		"""
		if fn not in self.on_exit_cbs:
			self.rescan_cbs.append(fn)

	def _set_profile(self, mapper, filename):
		# Called from socket server thread
		p = Profile(TalkingActionParser())
		p.load(filename).compress()
		self.profile_file = filename

		if mapper.profile.gyro and not p.gyro:
			# Turn off gyro sensor that was enabled but is no longer needed
			if mapper.get_controller():
				log.debug("Turning gyrosensor OFF")
				mapper.get_controller().set_gyro_enabled(False)
		elif not mapper.profile.gyro and p.gyro:
			# Turn on gyro sensor that was turned off, if profile has gyro action set
			if mapper.get_controller():
				log.debug("Turning gyrosensor ON")
				mapper.get_controller().set_gyro_enabled(True)
		# Cancel everything
		mapper.cancel_all()
		# Release all buttons
		mapper.release_virtual_buttons()
		# Reset mouse (issue #222)
		mapper.mouse.reset()

		# This last line kinda depends on GIL...
		mapper.profile = p
		# Re-apply all locks
		for c in self.clients:
			c.reaply_locks(self, mapper)
		if mapper.get_controller():
			self.send_profile_info(mapper.get_controller(), self._send_to_all)
		else:
			self.send_profile_info(None, self._send_to_all, mapper=mapper)

	def _send_to_all(self, message_str):
		"""Sends message to all connect clients.
		Should be called while lock is acquired.
		Message should be utf-8 encoded str.
		"""
		for client in self.clients:
			try:
				client.wfile.write(message_str)
			except Exception:
				pass

	def on_sa_turnoff(self, mapper, action):
		"""Called when 'turnoff' action is used"""
		if mapper.get_controller():
			mapper.get_controller().turnoff()

	def on_sa_restart(self, *a):
		"""Called when 'restart' action is used"""
		with self.lock:
			for c in self.clients:
				c.close()
		os.system("%s %s None restart &" % (sys.executable, sys.argv[0]))

	def on_sa_led(self, mapper, action):
		"""Called when 'led' action is used"""
		if mapper.get_controller():
			mapper.get_controller().set_led_level(action.brightness)

	def on_sa_shell(self, mapper, action):
		"""Called when 'shell' action is used"""
		return subprocess.Popen(action.command, shell=True)

	def on_sa_gestures(self, mapper, action, x, y, what):
		"""Called when 'gestures' action is used"""
		# TODO: Take up_direction from action
		gd = None
		with self.lock:
			if action.osd_enabled and self.osd_daemon:
				# When OSD is enabled, gesture detection is handled
				# by scc-osd-daemon.
				self.osd_daemon.gesture_action = action
				self._osd("gesture", "--controller", mapper.get_controller().get_id(), "--control-with", what)
				log.debug("Gesture detection request sent to scc-osd-daemon")
			else:
				# Otherwise it is handled internally
				up_direction = 0
				gd = self._start_gesture(
					mapper, what, up_direction, lambda gesture_string: action.gesture(mapper, gesture_string),
				)
		if gd:
			gd.enable()
			log.debug("Gesture detection started on %s", what)
			gd.whole(mapper, x, y, what)

	def on_sa_cemuhook(self, mapper, action, data):
		"""Called by 'cemuhook' action"""
		if self.cemuhook is None:
			try:
				self.cemuhook = CemuhookServer(self)
			except Exception as e:
				log.error("Failed to initialize CemuHookUDP Motion Provider: %s", e)
				return
		self.cemuhook.feed(data)

	def _osd(self, *data):
		"""Has to be called with self.lock held.
		Returns True on success.
		"""
		# Pre-format data
		data = b"OSD: %s\n" % (shjoin(data),)

		# Check if scc-osd-daemon is available
		if not self.osd_daemon:
			log.warning("Cannot show OSD; there is no scc-osd-daemon registered")
			return False
		# Send request
		try:
			self.osd_daemon.wfile.write(data)
			self.osd_daemon.wfile.flush()
		except Exception as e:
			log.error("Failed to display OSD: %s", e)
			self.osd_daemon = None
			return False
		return True

	def on_sa_osd(self, mapper, action):
		"""Called when 'osd' action is used"""
		with self.lock:
			self._osd("message", "-t", str(action.timeout), "-s", str(action.size), action.text)

	def on_sa_clearosd(self, mapper, action):
		"""Called when 'clearosd' action is used"""
		with self.lock:
			self._osd("clear")

	def on_sa_area(self, mapper, action, x1, y1, x2, y2):
		"""Called when *AreaAction has OSD enabled"""
		with self.lock:
			self._osd("area", "-x", str(x1), "-y", str(y1), "--width", str(x2 - x1), "--height", str(y2 - y1))

	def on_sa_clear_osd(self, *a):
		with self.lock:
			self._osd("clear")

	def on_sa_keyboard(self, mapper, action):
		"""Called when 'keyboard' action is used"""
		with self.lock:
			self._osd("keyboard")

	def on_sa_menu(self, mapper, action, *pars):
		"""Called when 'menu' action is used"""
		p = [action.MENU_TYPE]
		if mapper.get_controller():
			p += ["--controller", mapper.get_controller().get_id()]
		if "." in action.menu_id:
			path = find_menu(action.menu_id)
			if not path:
				log.error("Cannot show menu: Menu '%s' not found", action.menu_id)
				return
			p += ["--from-file", path]
		else:
			p += ["--from-profile", mapper.profile.get_filename(), action.menu_id]
		p += list(pars)

		with self.lock:
			self._osd(*p)

	on_sa_gridmenu = on_sa_menu

	def on_sa_dialog(self, mapper, action, *pars):
		# Replace actions with id, title pairs
		data = []
		self.osd_ids = {}
		for x in pars:
			if isinstance(x, Action):
				id = str(hash(x))
				self.osd_ids[id] = x.strip()
				data += [id, x.describe(Action.AC_MENU)]
			else:
				data.append(x)

		with self.lock:
			self._osd("dialog", *data)

	def on_sa_profile(self, mapper, action):
		"""Called when 'profile' action is used"""
		name = action.profile
		if "/" in name:
			# Small sanity check
			log.error("Cannot load profile: Profile '%s' not found", name)
			return
		path = find_profile(name)
		if path:
			with self.lock:
				try:
					self._set_profile(mapper, path)
					log.info("Loaded profile '%s'", name)
				except Exception as e:
					log.exception(e)
			return
		log.error("Cannot load profile: Profile '%s' not found", name)

	def on_start(self):
		os.chdir(self.cwd)

	def on_controller_status(self, sc, onoff):
		if onoff:
			log.debug("Controller turned ON")
		else:
			log.debug("Controller turned OFF")

	def sigterm(self, *a):
		self.exiting = True
		for fn in self.on_exit_cbs:
			fn(self)
		for d in (self.osd_daemon, self.autoswitch_daemon):
			if d:
				d.wfile.close()
		self.osd_daemon, self.autoswitch_daemon = None, None
		for p in self.subprocs:
			p.kill()
		self.subprocs = []
		sys.exit(0)

	def connect_x(self):
		"""Creates connection to X Server"""
		if "WAYLAND_DISPLAY" in os.environ:
			import gi

			gi.require_version("Gtk", "3.0")
			try:
				gi.require_version("GtkLayerShell", "0.1")
				from gi.repository import GtkLayerShell

				if GtkLayerShell.is_supported():
					log.warning(
						"Wayland detected. Disabling X11 support, some functionality will be unavailable",
					)
					self.xdisplay = None
					self.subprocs.append(Subprocess("scc-osd-daemon", True))
					return
			except (ImportError, ValueError):
				pass

			log.warning("Wayland without zwlr_layer_shell_v1 support detected! Disabling X11 support, some functionality, including OSD, will be unavailable")
			self.xdisplay = None
			return

		if "DISPLAY" not in os.environ:
			log.warning("DISPLAY env variable not set. Some functionality will be unavailable")
			self.xdisplay = None
			return

		self.xdisplay = X.open_display(os.environ["DISPLAY"].encode("utf-8"))
		if self.xdisplay:
			log.debug("Connected to XServer %s", os.environ["DISPLAY"])

			for c in self.controllers:
				if c.get_mapper():
					c.get_mapper().set_xdisplay(self.xdisplay)
			for m in self.free_mappers:
				m.set_xdisplay(self.xdisplay)
			if not self.alone:
				self.subprocs.append(Subprocess("scc-osd-daemon", True))
				if len(Config()["autoswitch"]):
					# Start scc-autoswitch-daemon only if there are some switch rules defined
					self.subprocs.append(Subprocess("scc-autoswitch-daemon", True))
		else:
			log.warning("Failed to connect to XServer. Some functionality will be unavailable")
			self.xdisplay = None

	def init_mapper(self):
		"""Setups new mapper instance.
		"""
		try:
			mapper = Mapper(Profile(TalkingActionParser()), self.scheduler, poller=self.poller)
		except CannotCreateUInputException as e:
			# Most likely UInput is not available
			# Create mapper with all virtual devices set to Dummies.
			log.exception(e)
			self.add_error("uinput", str(e))
			mapper = Mapper(Profile(TalkingActionParser()), self.scheduler, keyboard=None, mouse=None, gamepad=False)

		mapper.set_special_actions_handler(self)
		mapper.set_xdisplay(self.xdisplay)
		mapper.schedule(1.0, self.fix_xinput)
		return mapper

	def fix_xinput(self, mapper):
		name = mapper.get_gamepad_name()
		if self.xdisplay and Config()["fix_xinput"] and name:
			# Three conditions: X has to be available, 'fix_xinput' must
			# be enabled in config and controller should not be dummy
			# (should have a name)
			try:
				for d in xinput.get_devices():
					if d.get_name() == name:
						if d.is_pointer() and d.is_slave():
							d.float()
			except OSError as e:
				# Most likely 'xinput' executable not found
				log.warning("Failed to deatach gamepad from xinput master: %s", e)

	def load_default_profile(self, mapper=None):
		mapper = mapper or self.default_mapper
		if self.default_profile is None:
			try:
				self.default_profile = find_profile(Config()["recent_profiles"][0])
			except Exception:
				# Broken config is not reason to fail here
				pass
		try:
			mapper.profile.load(self.default_profile).compress()
		except Exception as e:
			log.warning("Failed to load profile. Starting with no mappings.")
			log.warning("Reason: %s", e)

	def add_controller(self, c):
		if len(self.free_mappers) > 0:
			# Reuse already created mapper, so SCC will not spam system
			# with fake devices
			mapper, self.free_mappers = self.free_mappers[0], self.free_mappers[1:]
			if mapper != self.default_mapper:
				log.debug("Reused mapper %s for %s", mapper, c)
		else:
			# New controller, but no mapper created
			mapper = self.init_mapper()
			self.load_default_profile(mapper)
		mapper.set_controller(c)
		c.set_mapper(mapper)
		if mapper == self.default_mapper:
			log.debug("Assigned default_mapper to %s", c)
		if mapper.profile.gyro:
			log.debug("Turning gyrosensor ON")
			c.set_gyro_enabled(True)

		c.apply_config(Config().get_controller_config(c.get_id()))
		self.controllers.append(c)
		log.debug("Controller added: %s", c)
		with self.lock:
			self.send_controller_list(self._send_to_all)
			self.send_all_profiles(self._send_to_all)

	def remove_controller(self, c):
		mapper = c.mapper
		if mapper:
			mapper.release_virtual_buttons()
		c.disconnected()

		with self.lock:
			while c in self.controllers:
				self.controllers.remove(c)
			log.debug("Controller removed: %s", c)

			if mapper == self.default_mapper and len(self.controllers) > 0:
				# Special case, default_mapper should be always
				# assigned to something, so if controller with default_mapper
				# is disconnected, it's reassigned to next available controller
				swap_c = self.controllers[0]
				swap_mapper = swap_c.get_mapper()
				swap_mapper.set_controller(None)
				swap_c.set_mapper(mapper)
				mapper.set_controller(swap_c)
				self.free_mappers.append(swap_mapper)
				log.debug("Reassigned default_mapper to %s", swap_c)
			else:
				c.set_mapper(None)
				if mapper:
					mapper.set_controller(None)
					self.free_mappers.append(mapper)
			self.send_controller_list(self._send_to_all)

	def get_active_ids(self):
		"""Returns iterable with IDs of all active controllers"""
		return [x.get_id() for x in self.controllers]

	def add_error(self, id, error):
		"""Adds error (string) to report. Used when USB driver reports that device
		cannot be accessed or when UInput is not available.

		Every error has id that can be later used to remove it from list to
		indicate that error has been resolved.
		"""
		with self.lock:
			self.errors.append((id, error))
			self._send_to_all(("Error: %s\n" % (error,)).encode("utf-8"))

	def remove_error(self, id):
		"""Removes error added with 'add_error'. If such error cannot be found,
		does nothing.

		When last error is removed, this method automatically sends "Ready."
		message to indicate that daemon is ready to serve clients.
		"""
		with self.lock:
			self.errors = [(_id, error) for (_id, error) in self.errors if _id != id]
			if len(self.errors) == 0:
				self._send_to_all(b"Ready.\n")

	def send_controller_list(self, method):
		"""Sends controller count and list of controllers using provided method
		"""
		for c in self.controllers:
			method(
				("Controller: %s %s %s %s\n" % (c.get_id(), c.get_type(), c.flags, c.get_gui_config_file())).encode(
					"utf-8",
				),
			)
		method(("Controller Count: %s\n" % (len(self.controllers),)).encode("utf-8"))

	def send_profile_info(self, controller, method, mapper=None):
		"""Sends info about current profile using provided method.
		Returns True if mapper is default_mapper.
		"""
		mapper = mapper or controller.mapper
		if controller:
			method(
				("Controller profile: %s %s\n" % (controller.get_id(), mapper.profile.get_filename())).encode("utf-8"),
			)
		if mapper == self.default_mapper:
			method(("Current profile: %s\n" % (mapper.profile.get_filename(),)).encode("utf-8"))
			return True
		return False

	def send_all_profiles(self, method):
		"""Sends info about all profiles assigned to all
		controllers using provided method.
		"""
		# As special case, at least default_mapper profile has to be sent always
		default_sent = False
		for c in self.controllers:
			default_sent = self.send_profile_info(c, method) or default_sent
		if not default_sent:
			self.send_profile_info(None, method, mapper=self.default_mapper)

	def run(self):
		log.debug("Starting SCCDaemon...")
		signal.signal(signal.SIGTERM, self.sigterm)
		self.init_drivers()
		self.dev_monitor.start()
		load_custom_module(log)
		self.default_mapper = self.init_default_mapper()
		self.free_mappers.append(self.default_mapper)
		self.load_default_profile()
		self.lock.acquire()
		self.start_listening()
		self.connect_x()
		self.lock.release()
		self.start_drivers()
		self.dev_monitor.rescan()

		while True:
			for fn in self.mainloops:
				fn()

	def start_listening(self):
		if os.path.exists(self.socket_file):
			os.unlink(self.socket_file)
		instance = self

		class SSHandler(StreamRequestHandler):
			def handle(self):
				instance._sshandler(self.connection, self.rfile, self.wfile)

		self.sserver = ThreadingUnixStreamServer(self.socket_file, SSHandler)
		t = threading.Thread(target=self.sserver.serve_forever)
		t.daemon = True
		t.start()
		os.chmod(self.socket_file, stat.S_IRUSR | stat.S_IWUSR)
		log.debug("Created control socket %s", self.socket_file)

	def _start_gesture(self, mapper, what, up_angle, callback):
		"""Starts gesture detection on specified pad.
		Calls callback with gesture string when finished.

		Should be called with lock held.
		"""
		gd = None

		def cb(detector, gesture):
			# This callback is expected to be called with lock held
			with self.lock:
				self._apply(mapper, what, lambda a: a.original_action)
			log.debug("Gesture detected on %s: %s", what, gesture)
			callback(gesture)

		def set(action):
			# ObservingAction should be above GestureDetector
			if isinstance(action, ObservingAction):
				gd.original_action = action.original_action
				action.original_action = gd
				return action
			gd.original_action = action
			return gd

		gd = GestureDetector(up_angle, cb)
		self._apply(mapper, what, set)
		return gd

	def _sshandler(self, connection, rfile, wfile):
		with self.lock:
			client = Client(connection, self.default_mapper, rfile, wfile)
			self.clients.add(client)
			wfile.write(b"SCCDaemon\n")
			wfile.write(("Version: %s\n" % (DAEMON_VERSION,)).encode("utf-8"))
			wfile.write(("PID: %s\n" % (os.getpid(),)).encode("utf-8"))
			self.send_controller_list(wfile.write)
			self.send_all_profiles(wfile.write)
			if len(self.errors) == 0:
				wfile.write(b"Ready.\n")
			else:
				for id, error in self.errors:
					wfile.write(("Error: %s\n" % (error,)).encode("utf-8"))

		while True:
			try:
				line = rfile.readline()
			except Exception:
				# Connection terminated
				break
			if len(line) == 0:
				break
			if len(line.strip(b"\t\n ")) > 0:
				self._handle_message(client, line.strip(b"\n"))

		with self.lock:
			client.unlock_actions(self)
			if self.osd_daemon == client:
				log.info("scc-osd-daemon lost")
				self.osd_daemon = None
			if self.autoswitch_daemon == client:
				log.info("scc-autoswitch-daemon lost")
				self.autoswitch_daemon = None
			self.clients.remove(client)

	def _handle_message(self, client, message: bytes):
		"""Handles message received from client."""
		if message.startswith(b"Profile:"):
			with self.lock:
				try:
					filename = message[8:].decode("utf-8").strip("\t ")
					self._set_profile(client.mapper, filename)
					log.info("Loaded profile '%s'", filename)
					client.wfile.write(b"OK.\n")
				except Exception as e:
					exc = traceback.format_exc()
					log.exception(e)
					tb = str(exc).encode("utf-8").decode("unicode_escape").encode("latin1")
					client.wfile.write(b"Fail: " + tb + b"\n")
		elif message.startswith(b"OSD:"):
			if not self.osd_daemon:
				client.wfile.write(b"Fail: Cannot show OSD; there is no scc-osd-daemon registered\n")
			else:
				try:
					text = message[5:].decode("utf-8").strip("\t ")
					with self.lock:
						if not self._osd("message", text):
							raise Exception
					client.wfile.write(b"OK.\n")
				except Exception:
					client.wfile.write(b"Fail: cannot display OSD\n")
		elif message.startswith(b"Feedback:"):
			try:
				position, amplitude = message[9:].strip().split(b" ", 2)
				data = HapticData(getattr(HapticPos, position.decode("ascii")), int(amplitude))
				if client.mapper.get_controller():
					client.mapper.get_controller().feedback(data)
				client.wfile.write(b"OK.\n")
			except Exception as e:
				log.exception(e)
				client.wfile.write((f"Fail: {e}\n").encode())
		elif message.startswith(b"Controller."):
			with self.lock:
				client.mapper = self.default_mapper
				client.wfile.write(b"OK.\n")
		elif message.startswith(b"Controller:"):
			with self.lock:
				try:
					controller_id = message[11:].strip().decode("utf-8")
					for c in self.controllers:
						if c.get_id() == controller_id:
							client.mapper = c.get_mapper()
							client.wfile.write(b"OK.\n")
							break
					else:
						raise Exception("goto fail")
				except Exception:
					client.wfile.write(b"Fail: no such controller\n")
		elif message.startswith(b"State."):
			if Config()["enable_sniffing"]:
				client.wfile.write(b"State: %s\n" % (str(client.mapper.state),))
			else:
				log.warning("Refused 'State' request: Sniffing disabled")
				client.wfile.write(b"Fail: Sniffing disabled.\n")
		elif message.startswith(b"Led:"):
			try:
				number = int(message[4:])
				number = clamp(0, number, 100)
			except Exception as e:
				client.wfile.write(f"Fail: {e}\n".encode())
				return
			if client.mapper.get_controller():
				client.mapper.get_controller().set_led_level(number)
		elif message.startswith(b"Observe:"):
			if Config()["enable_sniffing"]:
				to_observe = [x for x in message.split(b":", 1)[1].strip(b" \t\r").split(b" ")]
				with self.lock:
					for l in to_observe:
						client.observe_action(self, SCCDaemon.source_to_constant(l))
					client.wfile.write(b"OK.\n")
			else:
				log.warning("Refused 'Observe' request: Sniffing disabled")
				client.wfile.write(b"Fail: Sniffing disabled.\n")
		elif message.startswith(b"Replace:"):
			try:
				l, actionstr = message.split(b":", 1)[1].strip(b" \t\r").split(b" ", 1)
				action = TalkingActionParser().restart(actionstr.decode()).parse().compress()
			except Exception as e:
				e = str(e).encode("utf-8").decode("unicode_escape").encode("latin1")
				client.wfile.write(b"Fail: failed to parse: " + e + b"\n")
				return
			with self.lock:
				try:
					if not self._can_lock_action(client.mapper, SCCDaemon.source_to_constant(l)):
						client.wfile.write(b"Fail: Cannot lock " + l + b"\n")
						return
				except ValueError:
					tb = str(traceback.format_exc()).encode("utf-8").decode("unicode_escape").encode("latin1")
					client.wfile.write(b"Fail: " + tb + b"\n")
					return
				client.replace_action(self, SCCDaemon.source_to_constant(l), action)
				client.wfile.write(b"OK.\n")
		elif message.startswith(b"Lock:"):
			to_lock = [x for x in message.split(b":", 1)[1].strip(b" \t\r").split(b" ")]
			with self.lock:
				try:
					for l in to_lock:
						if not self._can_lock_action(client.mapper, SCCDaemon.source_to_constant(l)):
							client.wfile.write(b"Fail: Cannot lock " + l + b"\n")
							return
				except ValueError:
					tb = str(traceback.format_exc()).encode("utf-8").decode("unicode_escape").encode("latin1")
					client.wfile.write(b"Fail: " + tb + b"\n")
					return
				for l in to_lock:
					client.lock_action(self, SCCDaemon.source_to_constant(l))
				client.wfile.write(b"OK.\n")
		elif message.startswith(b"Unlock."):
			with self.lock:
				client.unlock_actions(self)
				client.wfile.write(b"OK.\n")
		elif message.startswith(b"Reconfigure."):
			with self.lock:
				# Load config
				cfg = Config()
				# Reconfigure connected controllers
				for c in self.controllers:
					c.apply_config(cfg.get_controller_config(c.get_id()))
				# Start or stop scc-autoswitch-daemon as needed
				need_autoswitch_daemon = len(cfg["autoswitch"]) > 0
				if need_autoswitch_daemon and self.xdisplay and not self.autoswitch_daemon:
					self.subprocs.append(Subprocess("scc-autoswitch-daemon", True))
				elif not need_autoswitch_daemon and self.autoswitch_daemon:
					self._remove_subproccess("scc-autoswitch-daemon")
					self.autoswitch_daemon.close()
					self.autoswitch_daemon = None
				# Respond
				try:
					client.wfile.write(b"OK.\n")
					self._send_to_all(b"Reconfigured.\n")
				except Exception:
					pass
		elif message.startswith(b"Rescan."):
			cbs = []
			with self.lock:
				cbs += self.rescan_cbs
				# Respond first
				try:
					client.wfile.write(b"OK.\n")
				except Exception:
					pass
			# Do stuff later
			# (this cannot be done while self.lock is held, as creating new
			# controller would create race condition)
			for cb in self.rescan_cbs:
				try:
					cb()
				except Exception as e:
					log.exception(e)
			# dev_monitor rescan has to be last to run
			try:
				self.dev_monitor.rescan()
			except Exception as e:
				log.exception(e)

		elif message.startswith(b"Turnoff."):
			to_turn_off = []
			with self.lock:
				if client.mapper.get_controller():
					to_turn_off.append(client.mapper.get_controller())
				else:
					to_turn_off += [c for c in self.controllers]
			for c in to_turn_off:
				c.turnoff()
			client.wfile.write(b"OK.\n")
		elif message.startswith(b"Gesture:"):
			try:
				what, up_angle = message[8:].strip().split(b" ", 2)
				up_angle = int(up_angle)
			except Exception:
				tb = str(traceback.format_exc()).encode("utf-8").decode("unicode_escape").encode("latin1")
				client.wfile.write(b"Fail: " + tb + b"\n")
				return
			with self.lock:
				client.request_gesture(self, what, up_angle)
				client.wfile.write(b"OK.\n")
		elif message.startswith(b"Restart."):
			self.on_sa_restart()
		elif message.startswith(b"Gestured:"):
			gstr = message[9:].strip()
			client.gesture_action.gesture(client.mapper, gstr)
			with self.lock:
				client.wfile.write(b"OK.\n")
		elif message.startswith(b"Selected:"):
			menuaction = None

			def press(mapper):
				try:
					menuaction.button_press(mapper)
					client.mapper.schedule(0.1, release)
				except Exception as e:
					log.error("Error while processing menu action")
					log.exception(e)

			def release(mapper):
				try:
					menuaction.button_release(mapper)
				except Exception as e:
					log.error("Error while processing menu action")
					log.exception(e)

			with self.lock:
				try:
					# menu_id, item_id = shsplit(message)[1:]
					# print(message)
					tmp = shsplit(str(message, "utf-8"))
					menu_id, item_id = tmp[1:]
					menuaction = None
					if menu_id in (None, "None"):
						menuaction = self.osd_ids[item_id]
					elif "." in menu_id:
						# TODO: Move this common place
						data = json.loads(open(menu_id).read())
						menudata = MenuData.from_json_data(data, TalkingActionParser())
						menuaction = menudata.get_by_id(item_id).action
					else:
						menuaction = client.mapper.profile.menus[menu_id].get_by_id(item_id).action
					client.wfile.write(b"OK.\n")
				except Exception:
					log.warning("Selected menu item is no longer valid.")
					client.wfile.write(b"Fail: Selected menu item is no longer valid\n")
				if menuaction:
					client.mapper.schedule(0, press)
		elif message.startswith(b"Register:"):
			with self.lock:
				if message.strip().endswith(b"osd"):
					if self.osd_daemon:
						self.osd_daemon.close()
					self.osd_daemon = client
					log.info("Registered scc-osd-daemon")
				elif message.strip().endswith(b"autoswitch"):
					if self.autoswitch_daemon:
						self.autoswitch_daemon.close()
					self.autoswitch_daemon = client
					log.info("Registered scc-autoswitch-daemon")
				client.wfile.write(b"OK.\n")
		else:
			client.wfile.write(b"Fail: Unknown command\n")

	def _remove_subproccess(self, binary_name):
		"""Removes subproccess started with specified binary name from list of
		managed subproccesses, effectively preventing daemon from
		auto-restarting it.

		Should be called while lock is acquired
		"""
		n = []
		for i in self.subprocs:
			if i.binary_name == binary_name:
				i.mark_killed()
			else:
				n.append(i)
		self.subprocs = n

	def _can_lock_action(self, mapper, what):
		"""Returns True if action assigned to axis,
		pad or button is not yet locked.

		Should be called while self.lock is acquired.
		"""
		# TODO: Probably move to mapper
		is_locked = lambda a: isinstance(a, LockedAction) or (
			isinstance(a, ObservingAction) and isinstance(a.original_action, LockedAction)
		)

		if what == STICK:
			if is_locked(mapper.profile.buttons[SCButtons.STICKPRESS]):
				return False
			if is_locked(mapper.profile.stick):
				return False
			return True
		if what == RSTICK:
			if is_locked(mapper.profile.buttons[SCButtons.RSTICKPRESS]):
				return False
			if is_locked(mapper.profile.rstick):
				return False
			return True
		if what == SCButtons.LT:
			return not is_locked(mapper.profile.triggers[LEFT])
		if what == SCButtons.RT:
			return not is_locked(mapper.profile.triggers[RIGHT])
		if what in SCButtons.__members__.values():
			return not is_locked(mapper.profile.buttons[what])
		if what in (LEFT, RIGHT, CPAD, DPAD):
			return not is_locked(mapper.profile.pads[what])
		return False

	def _apply(self, mapper, what, callback, *args):
		"""Applies callback on action that is currently set to input specified
		by 'what'. Raises ValueError if what is not known.

		For example, if what == STICK, executes
			mapper.profile.stick = callback(mapper.profile.stick, *args)
		"""
		if what == STICK:
			mapper.profile.stick = callback(mapper.profile.stick, *args)
		elif what == RSTICK:
			mapper.profile.rstick = callback(mapper.profile.rstick, *args)
		elif what == SCButtons.LT:
			mapper.profile.triggers[LEFT] = callback(mapper.profile.triggers[LEFT], *args)
		elif what == SCButtons.RT:
			mapper.profile.triggers[RIGHT] = callback(mapper.profile.triggers[RIGHT], *args)
		elif what in SCButtons.__members__.values():
			r = callback(mapper.profile.buttons[what], *args)
			mapper.profile.buttons[what] = r
		elif what in (LEFT, RIGHT):
			if what == LEFT:
				mapper.buttons &= ~SCButtons.LPADTOUCH
			else:
				mapper.buttons &= ~SCButtons.RPADTOUCH
			a = callback(mapper.profile.pads[what], *args)
			a.whole(mapper, 0, 0, what)
			mapper.profile.pads[what] = a
		elif what in (CPAD, DPAD):
			a = callback(mapper.profile.pads[what], *args)
			a.whole(mapper, 0, 0, what)
			mapper.profile.pads[what] = a
		else:
			raise ValueError("Unknown source: %s" % (what,))

	@staticmethod
	def source_to_constant(s):
		"""Turns string as 'A', 'LEFT' or 'ABS_X' into one of SCButtons.*,
		LEFT, RIGHT or STICK constants.

		Raises ValueError if passed string cannot be converted.

		Used when parsing `Lock: ...` message
		"""
		s = s.decode("utf-8").strip(" \t\r\n")
		if s in (STICK, LEFT, RIGHT, CPAD):
			return s
		if s == "STICKPRESS":
			# Special case, as that button is actually named STICK :(
			return SCButtons.STICKPRESS
		if hasattr(SCButtons, s):
			return getattr(SCButtons, s)
		raise ValueError("Unknown source: %s" % (s,))

	def _remove_socket(self):
		self.sserver.shutdown()
		if os.path.exists(self.socket_file):
			os.unlink(self.socket_file)
		log.debug("Control socket removed")

	def debug(self):
		set_logging_level(True, True)
		self.on_start()
		self.write_pid()
		try:
			self.run()
		except KeyboardInterrupt:
			log.debug("Break")
		self.sigterm()


class Client:
	def __init__(self, connection, mapper, rfile, wfile):
		self.connection = connection
		self.rfile = rfile
		self.wfile = wfile
		self.mapper = mapper
		self.gesture_action = None
		self.locked_actions = {}

	def close(self):
		"""Closes connection to this client"""
		try:
			self.connection.shutdown(True)
		except Exception:
			pass

	def request_gesture(self, daemon, what, up_angle):
		"""Handler used when client requested gesture detection with
		"Gesture:" message.

		Should be called while daemon.lock is acquired.
		"""

		def cb(gesture):
			# Called while lock is being held
			try:
				self.wfile.write(b"Gesture: %s %s\n" % (what, gesture))
			except Exception:
				pass

		gd = daemon._start_gesture(self.mapper, what, up_angle, cb)
		gd.enable()
		log.debug("Gesture detection requested on %s", what)

	def lock_action(self, daemon, what):
		"""Locks action so event can be send to client instead of handling it.

		Should be called while daemon.lock is acquired.
		"""

		def lock(action, what):
			# ObservingAction should be above LockedAction
			if isinstance(action, ObservingAction):
				action.original_action = LockedAction(what, self, action.original_action)
				return action
			return LockedAction(what, self, action)

		daemon._apply(self.mapper, what, lock, what)

	def observe_action(self, daemon, what):
		"""Enables observing of action so event is both sent to client and handled.

		Should be called while daemon.lock is acquired.
		"""
		daemon._apply(self.mapper, what, lambda a: ObservingAction(what, self, a))

	def replace_action(self, daemon, what, action):
		"""Temporally replaces action in way that allows reversing operation when
		client disconnects.

		Should be called while daemon.lock is acquired.
		"""
		daemon._apply(self.mapper, what, lambda a: ReplacedAction(what, self, action, a))

	def unlock_actions(self, daemon):
		"""Should be called while daemon.lock is acquired"""
		locked, self.locked_actions = self.locked_actions, {}
		for mapper in locked:
			s = locked[mapper]
			for a in s:
				a.unlock(daemon)

	def reaply_locks(self, daemon, mapper):
		"""Called after profile is changed.
		Should be called while daemon.lock is acquired
		"""
		if mapper in self.locked_actions:
			s, self.locked_actions[mapper] = self.locked_actions[mapper], set()
			for a in s:
				a.reaply(self, daemon)


class ReportingAction(Action):
	"""Action used to send requested inputs to client.
	Base for LockedAction and ObservingAction
	"""

	MIN_DIFFERENCE = 300

	def __init__(self, what, client):
		self.what = what
		self.client = client
		self.mapper = client.mapper
		self.old_pos = 0, 0

	def _store_lock(self):
		if self.mapper not in self.client.locked_actions:
			self.client.locked_actions[self.mapper] = set()
		self.client.locked_actions[self.mapper].add(self)

	def __repr__(self):
		return "<%s of %x>" % (self.__class__.__name__, hash(self.client))

	__str__ = __repr__

	def _report(self, message):
		try:
			self.client.wfile.write(message.encode("utf-8"))
		except Exception:
			# May fail when client dies
			self.client.rfile.close()
			self.client.wfile.close()

	def trigger(self, mapper, position, old_position):
		if mapper.get_controller():
			self._report(
				"Event: %s %s %s %s\n" % (mapper.get_controller().get_id(), nameof(self.what), position, old_position),
			)

	def button_press(self, mapper, number=1):
		if mapper.get_controller():
			if self.what == SCButtons.STICKPRESS:
				self._report("Event: %s STICKPRESS %s\n" % (mapper.get_controller().get_id(), number))
			else:
				self._report("Event: %s %s %s\n" % (mapper.get_controller().get_id(), nameof(self.what), number))

	def button_release(self, mapper):
		ReportingAction.button_press(self, mapper, 0)

	def whole(self, mapper, x, y, what):
		min_difference = self.MIN_DIFFERENCE
		if what == CPAD:
			min_difference /= 10
		if x == 0 or y == 0 or abs(x - self.old_pos[0]) > min_difference or abs(y - self.old_pos[1] > min_difference):
			self.old_pos = x, y
			if mapper.get_controller():
				self._report("Event: %s %s %s %s\n" % (mapper.get_controller().get_id(), what, x, y))


class LockedAction(ReportingAction):
	"""Temporal action used to send requested inputs to client"""

	def __init__(self, what, client, original_action):
		ReportingAction.__init__(self, what, client)
		self.original_action = original_action
		original_action.cancel(self.mapper)
		self._store_lock()
		log.debug("%s locked by %s", nameof(self.what), self.client)

	def reaply(self, client, daemon):
		client.lock_action(daemon, self.what)

	def unlock(self, daemon):
		def _unlock(a):
			if isinstance(a, ObservingAction):
				# Needs to be handled specifically
				a.original_action = _unlock(a.original_action)
				return a
			if isinstance(a, LockedAction):
				return a.original_action
			return a

		daemon._apply(self.mapper, self.what, _unlock)
		log.debug("%s unlocked", nameof(self.what))


class ReplacedAction(LockedAction):
	def __init__(self, what, client, new_action, original_action):
		ReportingAction.__init__(self, what, client)
		self.original_action = original_action
		self.new_action = new_action.compress()
		original_action.cancel(self.mapper)
		self._store_lock()
		log.debug("%s replaced by %s", nameof(self.what), self.client)

	def reaply(self, client, daemon):
		client.replace_action(daemon, self.what, self.new_action)

	def trigger(self, mapper, position, old_position):
		self.new_action.trigger(mapper, position, old_position)

	def button_press(self, mapper, number=1):
		self.new_action.button_press(mapper, mapper)

	def button_release(self, mapper):
		self.new_action.button_release(mapper, mapper)

	def whole(self, mapper, x, y, what):
		self.new_action.whole(mapper, x, y, what)


class ObservingAction(ReportingAction):
	"""Similar to LockedAction, send inputs to client *and* executes actions.
	"""

	def __init__(self, what, client, original_action):
		ReportingAction.__init__(self, what, client)
		self.original_action = original_action
		self._store_lock()
		log.debug("%s on %s observed by %x", nameof(self.what), client.mapper.get_controller(), hash(self.client))

	def reaply(self, client, daemon):
		client.observe_action(daemon, self.what)

	def cancel(self, mapper):
		self.original_action.cancel(mapper)

	def unlock(self, daemon):
		def _unobserve(a):
			if isinstance(a, ObservingAction):
				if a.client == self.client:
					return a.original_action
				a.original_action = _unobserve(a.original_action)
				return a
			if isinstance(a, LockedAction):
				a.original_action = _unobserve(a.original_action)
				return a
			return a

		daemon._apply(self.mapper, self.what, _unobserve)
		log.debug("%s on %s no longer observed by %x", nameof(self.what), self.mapper.get_controller(), hash(self.client))

	def trigger(self, mapper, position, old_position):
		ReportingAction.trigger(self, mapper, position, old_position)
		self.original_action.trigger(mapper, position, old_position)

	def button_press(self, mapper, number=1):
		ReportingAction.button_press(self, mapper, number)
		self.original_action.button_press(mapper)

	def button_release(self, mapper):
		ReportingAction.button_release(self, mapper)
		self.original_action.button_release(mapper)

	def whole(self, mapper, x, y, what):
		ReportingAction.whole(self, mapper, x, y, what)
		self.original_action.whole(mapper, x, y, what)


class Subprocess:
	"""Part of scc-daemon executed as another process, killed along with scc-daemon.
	Currently scc-osd-daemon and scc-windowswitch-daemon.
	"""

	def __init__(self, binary_name, debug, restart_after=5):
		self.binary_name = binary_name
		self.restart_after = restart_after
		self.args = [sys.executable, find_binary(binary_name)]
		if debug:
			self.args.append("debug")
		self._killed = False
		self.p = None
		self.t = threading.Thread(target=self._threaded)
		self.t.daemon = True
		self.t.start()

	def _threaded(self, *a):
		while not self._killed:
			self.p = subprocess.Popen(self.args, stdin=None)
			self.p.communicate()
			if self.p and self.p.returncode == 8:
				log.warning("%s exited with code 8; not restarting", self.binary_name)
				self.p = None
				return
			self.p = None
			if not self._killed:
				log.warning("%s died; restarting after %ss", self.binary_name, self.restart_after)
				time.sleep(self.restart_after)

	def mark_killed(self):
		"""Prevents subprocess from being automatically restarted, but doesn't
		really kill it.
		"""
		self._killed = True

	def kill(self):
		self.mark_killed()
		if self.p:
			self.p.kill()
		self.p = None



"""SC Controller - Autoswitch Daemon

Observes active window and commands scc-daemon to change profiles as needed.
"""
from __future__ import annotations

import logging
import os
import re
import socket
import threading
import time

from scc.config import Config
from scc.lib import xwrappers as X
from scc.mapper import Mapper
from scc.menu_data import MENU_GENERATORS, MenuGenerator, MenuItem, Separator
from scc.parser import TalkingActionParser
from scc.paths import get_daemon_socket
from scc.special_actions import ChangeProfileAction
from scc.tools import _, find_profile

log = logging.getLogger("AutoSwitcher")


class AutoSwitcher:
	INTERVAL = 1

	def __init__(self):
		self.dpy = X.open_display(os.environ["DISPLAY"].encode("utf-8"))
		self.lock = threading.Lock()
		self.thread = threading.Thread(target=self.connect_daemon)
		self.config = Config()
		self.mapper = Mapper(None, None, keyboard=None, mouse=None, gamepad=None)
		self.mapper.set_special_actions_handler(self)
		self.enabled = False
		self.socket = None
		self.connected = False
		self.exit_code = None
		self.current_profile = None
		self.current_window = None
		self.conds = AutoSwitcher.parse_conditions(self.config)

	@staticmethod
	def parse_conditions(config):
		"""Parses conditions from config"""
		parser = TalkingActionParser()
		conds = {}
		for c in config["autoswitch"]:
			try:
				astr = c["action"]
				if type(astr) == dict and "action" in astr:
					# Backwards compatibility
					astr = astr["action"]
				action = parser.restart(astr).parse()
				conds[Condition.parse(c["condition"])] = action
			except Exception as e:
				# Failure here is not fatal
				log.error("Failed to parse autoswitcher condition '%s'", c)
				log.error(e)
		log.debug("Parsed %s autoswitcher conditions", len(conds))
		return conds

	@staticmethod
	def assign(conds, title, wm_class, profile):
		c = Condition(wm_class=wm_class[0])
		conds[c] = ChangeProfileAction(profile)

	@staticmethod
	def unassign(conds, title, wm_class, action):
		"""Removes any condition that matches given title/class/action combination.
		'action' can be None, in which case, removes removes any condition
		that matches title or wm class.
		"""
		count, cmpwith = 0, None
		if action is not None:
			cmpwith = action.to_string()
		for c in conds.copy().keys():
			if action is None or conds[c].to_string() == cmpwith:
				if c.matches(title, wm_class):
					del conds[c]
					count += 1
		log.debug("Removed %s autoswitcher conditions", count)

	def connect_daemon(self, *a) -> None:
		try:
			self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
			self.socket.connect(get_daemon_socket())
			self.socket.send(b"Register: autoswitch\n")
		except Exception:
			log.error("Failed to connect to scc-daemon")
			os._exit(1)
			return
		buffer = ""
		while self.exit_code is None:
			r = self.socket.recv(1024)
			self.lock.acquire()
			if len(r) == 0:
				self.lock.release()
				log.error("Connection to daemon lost")
				os._exit(2)
				return
			buffer += r.decode("utf-8")
			while "\n" in buffer:
				line, buffer = buffer.split("\n", 1)
				if line.startswith("Version:"):
					version = line.split(":", 1)[-1].strip()
					log.debug("Connected to daemon, version %s", version)
				elif line.startswith("Current profile:"):
					profile = line.split(":", 1)[-1].strip()
					log.debug("Daemon reported profile change: %s", profile)
					self.current_profile = profile
				elif line.startswith("Reconfigured."):
					log.debug("Reloading config...")
					self.config = Config()
					self.conds = AutoSwitcher.parse_conditions(self.config)
				elif line.startswith("Controller Count:"):
					self.enabled = int(line.split(":")[-1]) > 0
					log.debug("Enabled: %s", self.enabled)

			self.lock.release()

	def check(self, *a):
		w = X.get_current_window(self.dpy)
		if w == self.current_window or not self.current_profile:
			# Window not switched or profile is not known yet
			return
		self.current_window = w
		log.debug("Window switched: %s", w)
		pars = X.get_window_title(self.dpy, w), X.get_window_class(self.dpy, w)

		if pars[0] is None:
			pars = ("", pars[1])

		if pars[1] is None:
			pars = (pars[0], ("", ""))

		for c in self.conds:
			if c.matches(*pars):
				action = self.conds[c]
				action.button_press(self.mapper)
				action.button_release(self.mapper)

	def on_sa_profile(self, mapper, action):
		profile_name = action.profile
		path = find_profile(profile_name)
		if path:
			with self.lock:
				if path != self.current_profile and not self.current_profile.endswith(".mod"):
					# Switch only if target profile is not active
					# and active profile is not being editted.
					try:
						if self.config["autoswitch_osd"]:
							msg = _("Switched to profile") + " " + profile_name
							self.socket.send(b"OSD: " + msg.encode("utf-8") + b"\n")
						self.socket.send(b"Profile: " + path.encode("utf-8") + b"\n")
					except:
						log.error("Socket write failed")
						os._exit(2)
						return
		else:
			log.error("Cannot switch to profile '%s', profile file not found", profile_name)

	def on_sa_turnoff(self, mapper, action):
		with self.lock:
			try:
				self.socket.send(b"Turnoff.\n")
			except:
				log.error("Socket write failed")
				os._exit(2)

	def on_sa_restart(self, *a):
		with self.lock:
			try:
				self.socket.send(b"Restart.\n")
			except:
				log.error("Socket write failed")
				os._exit(2)

	def sigint(self, *a):
		log.error("break")
		os._exit(0)

	def run(self):
		self.thread.start()
		log.debug("AutoSwitcher started")
		while self.exit_code is None:
			if self.enabled:
				self.check()
			time.sleep(self.INTERVAL)
		return 1


class Condition:
	"""Represents AutoSwitcher condition loaded from configuration file.

	Currently, there are 4 ways to match window:
	By exact title, by part of title, by regexp aplied on title and by matching
	window class.
	It's possible to combine all three types of title matching with window class
	matching.
	"""

	def __init__(self, exact_title=None, title=None, regexp=None, wm_class=None):
		"""At least one parameter has to be specified; regexp has to be
		compiled regular expression.
		"""
		self.exact_title = exact_title
		self.title = title
		self.regexp = regexp
		if type(self.regexp) is str:
			self.regexp = re.compile(self.regexp)
		self.wm_class = wm_class
		self.empty = not (title or title or regexp or wm_class)

	def __str__(self):
		return "<Condition title=%s, exact_title=%s, regexp=%s, wm_class=%s>" % (
			self.title,
			self.exact_title,
			self.regexp,
			self.wm_class,
		)

	def describe(self):
		"""Returns string that describes condition in human-readable form.
		Used in GUI.
		"""
		rv = []
		if self.title:
			rv += [_("title contains '%s'") % (self.title,)]
		if self.exact_title:
			rv += [_("title is '%s'") % (self.exact_title,)]
		if self.regexp:
			rv += [_("title matches '%s'") % (self.regexp.pattern,)]
		if self.wm_class:
			rv += [_("class is '%s'") % (self.wm_class,)]
		if rv:
			return _(" and ").join(rv)
		return _("matches nothing")

	@staticmethod
	def parse(data) -> Condition:
		if "regexp" in data:
			data = dict(data)
			data["regexp"] = re.compile(data["regexp"])
		return Condition(**data)

	def encode(self):
		"""Returns Condition in dict that can be stored in json configuration"""
		rv = {}
		if self.title:
			rv["title"] = self.title
		if self.exact_title:
			rv["exact_title"] = self.exact_title
		if self.regexp:
			rv["regexp"] = self.regexp.pattern
		if self.wm_class:
			rv["wm_class"] = self.wm_class
		return rv

	def matches(self, window_title, wm_class):
		"""Returns True if condition matches provided window properties.

		wm_class is what xwrappers.get_window_class returns, tuple of two strings.
		"""
		if self.empty:
			# Empty condition matches nothing
			return False

		if self.wm_class:
			if self.wm_class != wm_class[0] and self.wm_class != wm_class[1]:
				# Window class matching is enabled and window doesn't match
				return False

		if self.exact_title and self.exact_title != window_title:
			# Matching exact title is enabled, but title doesn't match
			return False

		if self.title and self.title not in window_title:
			# Matching part of title is enabled, but doesn't match
			return False

		if self.regexp and not self.regexp.match(window_title):
			# Matching by regexp is enabled, but regexp doesn't match
			return False

		return True


class AutoswitchOptsMenuGenerator(MenuGenerator):
	"""Generates entire Autoswich Options submenu"""

	GENERATOR_NAME: str = "autoswitch"

	def callback(self, menu, daemon, controller, menuitem):
		def on_response(*a):
			menu.quit(-2)

		if menuitem.id in ("as::unassign", "as::assign"):
			if menuitem.id == "as::unassign":
				AutoSwitcher.unassign(self.conds, self.title, self.wm_class, self.assigned_prof)
			elif controller.get_profile():
				profile = os.path.split(controller.get_profile())[-1]
				if profile.endswith(".mod"):
					profile = profile[0:-4]
				if profile.endswith(".sccprofile"):
					profile = profile[0:-11]
				AutoSwitcher.unassign(self.conds, self.title, self.wm_class, None)
				AutoSwitcher.assign(self.conds, self.title, self.wm_class, profile)
			cfg = Config()
			cfg["autoswitch"] = [{"condition": c.encode(), "action": self.conds[c].to_string()} for c in self.conds]
			cfg.save()
			daemon.request(b"Reconfigure.\n", on_response, on_response)
		else:
			on_response()

	def describe(self):
		return _("[ All Profiles ]")

	def generate(self, menuhandler) -> list[MenuItem]:
		rv: list[MenuItem] = []
		if os.environ.get("WAYLAND_DISPLAY"):
			rv.append(self.mk_item(None, _("Autoswitch is not currently supported in Wayland")))
			return rv

		win = X.get_current_window(menuhandler.xdisplay)
		if not win:
			# Bail out if active window cannot be determined
			rv.append(self.mk_item(None, _("No active window")))
			rv.append(self.mk_item("as::close", _("Close")))
			return rv

		self.title = X.get_window_title(menuhandler.xdisplay, win)
		self.wm_class = X.get_window_class(menuhandler.xdisplay, win)
		self.assigned_prof = None
		self.conds = AutoSwitcher.parse_conditions(Config())
		if self.title and "-" in self.title:
			self.title = self.title.split("-")[-1]
		for c in self.conds:
			if c.matches(self.title, self.wm_class):
				self.assigned_prof = self.conds[c]
				break
		if win:
			display_title = self.title or _("No Title")
			rv.append(self.mk_item(None, _("Current Window: %s") % (self.title[0:25],)))
			if self.assigned_prof:
				rv.append(self.mk_item(None, _("Assigned Profile: %s") % (self.assigned_prof,)))
			else:
				rv.append(self.mk_item(None, _("No Profile Assigned")))
			rv.append(Separator())
			rv.append(Separator())
			rv.append(Separator())
			if self.assigned_prof:
				rv.append(self.mk_item("as::unassign", _("Unassign Profile")))
			rv.append(self.mk_item("as::assign", _("Assign Current Profile")))
		return rv

	def mk_item(self, item_id: str | None, title, **kws) -> MenuItem:
		"""Creates a MenuItem and assigns it a callback"""
		menuitem = MenuItem(item_id, title)
		menuitem.callback = self.callback
		for k in kws:
			setattr(menuitem, k, kws[k])
		return menuitem

def register_menu_generators() -> None:
	"""Add class to MENU_GENERATORS dict

	Needs to be called from menu otherwise autoswitcher entries will be empty
	"""
	MENU_GENERATORS[AutoswitchOptsMenuGenerator.GENERATOR_NAME] = AutoswitchOptsMenuGenerator



"""SCC - Steam Deck Driver

Based on sc_by_cable and steamdeck.c

Deck uses slightly different packed format and so common handle_inpu is not used.

On top of that, deck will automatically enable lizard mode unless requested
to not do so periodically.
"""

import ctypes
import logging
import struct
from enum import IntEnum

from scc.constants import STICK_PAD_MAX, STICK_PAD_MIN, ControllerFlags, SCButtons
from scc.drivers.sc_dongle import SCController, SCPacketType
from scc.drivers.usb import USBDevice, register_hotplug_device

VENDOR_ID = 0x28DE
PRODUCT_ID = 0x1205
ENDPOINT = 3
CONTROLIDX = 2
PACKET_SIZE = 128
UNLIZARD_INTERVAL = 100
# Basically, sticks on deck tend to return to non-zero position
STICK_DEADZONE = 3000

log = logging.getLogger("deck")


class DeckInput(ctypes.Structure):
	_fields_ = [
		("type", ctypes.c_uint8),
		("_a1", ctypes.c_uint8 * 3),
		("seq", ctypes.c_uint32),
		("buttons", ctypes.c_uint64),
		("lpad_x", ctypes.c_int16),
		("lpad_y", ctypes.c_int16),
		("rpad_x", ctypes.c_int16),
		("rpad_y", ctypes.c_int16),
		("accel_x", ctypes.c_int16),
		("accel_y", ctypes.c_int16),
		("accel_z", ctypes.c_int16),
		("gpitch", ctypes.c_int16),
		("groll", ctypes.c_int16),
		("gyaw", ctypes.c_int16),
		("q1", ctypes.c_uint16),
		("q2", ctypes.c_uint16),
		("q3", ctypes.c_uint16),
		("q4", ctypes.c_uint16),
		("ltrig", ctypes.c_uint16),
		("rtrig", ctypes.c_uint16),
		("stick_x", ctypes.c_int16),
		("stick_y", ctypes.c_int16),
		("rstick_x", ctypes.c_int16),
		("rstick_y", ctypes.c_int16),
		# Values above are readed directly from deck
		# Values bellow are converted so mapper can understand them
		("dpad_x", ctypes.c_int16),
		("dpad_y", ctypes.c_int16),
	]


class DeckButton(IntEnum):
	DOTS        = 0b100000000000000000000000000000000000000000000000000
	RSTICKTOUCH = 0b000100000000000000000000000000000000000000000000000
	LSTICKTOUCH = 0b000010000000000000000000000000000000000000000000000
	RGRIP2      = 0b000000001000000000000000000000000000000000000000000
	LGRIP2      = 0b000000000100000000000000000000000000000000000000000
	RSTICKPRESS = 0b000000000000000000000000100000000000000000000000000
	LSTICKPRESS = 0b000000000000000000000000000010000000000000000000000
	# bit 21 unused?
	RPADTOUCH   = 0b000000000000000000000000000000100000000000000000000
	LPADTOUCH   = 0b000000000000000000000000000000010000000000000000000
	RPADPRESS   = 0b000000000000000000000000000000001000000000000000000
	LPADPRESS   = 0b000000000000000000000000000000000100000000000000000
	RGRIP       = 0b000000000000000000000000000000000010000000000000000
	LGRIP       = 0b000000000000000000000000000000000001000000000000000
	START       = 0b000000000000000000000000000000000000100000000000000
	C           = 0b000000000000000000000000000000000000010000000000000
	BACK        = 0b000000000000000000000000000000000000001000000000000
	DPAD_DOWN   = 0b000000000000000000000000000000000000000100000000000
	DPAD_LEFT   = 0b000000000000000000000000000000000000000010000000000
	DPAD_RIGHT  = 0b000000000000000000000000000000000000000001000000000
	DPAD_UP     = 0b000000000000000000000000000000000000000000100000000
	A           = 0b000000000000000000000000000000000000000000010000000
	X           = 0b000000000000000000000000000000000000000000001000000
	B           = 0b000000000000000000000000000000000000000000000100000
	Y           = 0b000000000000000000000000000000000000000000000010000
	LB          = 0b000000000000000000000000000000000000000000000001000
	RB          = 0b000000000000000000000000000000000000000000000000100
	LT          = 0b000000000000000000000000000000000000000000000000010
	RT          = 0b000000000000000000000000000000000000000000000000001


DIRECTLY_TRANSLATABLE_BUTTONS = (
	0
	| DeckButton.A
	| DeckButton.B
	| DeckButton.X
	| DeckButton.Y
	| DeckButton.LB
	| DeckButton.RB
	| DeckButton.LT
	| DeckButton.RT
	| DeckButton.START
	| DeckButton.C
	| DeckButton.BACK
	| DeckButton.RGRIP
	| DeckButton.LGRIP
	| DeckButton.RPADTOUCH
	| DeckButton.LPADTOUCH
	| DeckButton.RPADPRESS
	| DeckButton.LPADPRESS
)


def map_button(i, from_, to):
	return to if (i.buttons & from_) else 0


def map_dpad(i, low, hi) -> int:
	if (i.buttons & low) != 0:
		return STICK_PAD_MIN
	if (i.buttons & hi) != 0:
		return STICK_PAD_MAX
	return 0


def apply_deadzone(value, deadzone):
	if value > -deadzone and value < deadzone:
		return 0
	return value


class Deck(USBDevice, SCController):
	flags = (
		0
		| ControllerFlags.SEPARATE_STICK
		| ControllerFlags.HAS_DPAD
		| ControllerFlags.IS_DECK
		| ControllerFlags.HAS_RSTICK
	)

	def __init__(self, device, handle, daemon):
		self.daemon = daemon
		USBDevice.__init__(self, device, handle)
		SCController.__init__(self, self, CONTROLIDX, ENDPOINT)
		self._old_state = DeckInput()
		self._input = DeckInput()
		self._ready = False

		self.claim_by(klass=3, subclass=0, protocol=0)
		self.read_serial()

	def generate_serial(self):
		self._serial = "%s:%s" % (self.device.getBusNumber(), self.device.getPortNumber())

	def disconnected(self):
		# Overrided to skip returning serial# to pool.
		pass

	def set_gyro_enabled(self, enabled):
		# Always on on deck
		pass

	def get_gyro_enabled(self) -> bool:
		# Always on on deck
		return True

	def get_type(self) -> str:
		return "deck"

	def __repr__(self) -> str:
		return f"<Deck {self.get_id()}>"

	def get_gui_config_file(self) -> str:
		return "deck.config.json"

	def configure(self, idle_timeout=None, enable_gyros=None, led_level=None) -> None:
		FORMAT = b">BBBB60x"
		# Timeout & Gyros
		self._driver.overwrite_control(self._ccidx, struct.pack(FORMAT, SCPacketType.CONFIGURE, 0x03, 0x08, 0x07))

	def clear_mappings(self):
		FORMAT = b">BB62x"
		# Timeout & Gyros
		self._driver.overwrite_control(self._ccidx, struct.pack(FORMAT, SCPacketType.CLEAR_MAPPINGS, 0x01))

	def on_serial_got(self):
		log.debug("Got SteamDeck with serial %s", self._serial)
		self._id = f"deck{self._serial}"
		self.set_input_interrupt(ENDPOINT, 64, self._on_input)

	def _on_input(self, endpoint: int, data: bytearray) -> None:
		if not self._ready:
			self.daemon.add_controller(self)
			self.configure()
			self._ready = True

		self._old_state, self._input = self._input, self._old_state
		ctypes.memmove(ctypes.addressof(self._input), bytes(data), len(data))
		if self._input.seq % UNLIZARD_INTERVAL == 0:
			# Keeps lizard mode from happening
			self.clear_mappings()

		# Handle dpad
		self._input.dpad_x = map_dpad(self._input, DeckButton.DPAD_LEFT, DeckButton.DPAD_RIGHT)
		self._input.dpad_y = map_dpad(self._input, DeckButton.DPAD_DOWN, DeckButton.DPAD_UP)
		# Convert buttons
		self._input.buttons = (
			0
			| ((self._input.buttons & DIRECTLY_TRANSLATABLE_BUTTONS) << 8)
			| map_button(self._input, DeckButton.DOTS, SCButtons.DOTS)
			# | map_button(self._input, DeckButton.RSTICKTOUCH, ....)	// not mapped
			# | map_button(self._input, DeckButton.LSTICKTOUCH, ....) // not mapped
			| map_button(self._input, DeckButton.LSTICKPRESS, SCButtons.STICKPRESS)
			| map_button(self._input, DeckButton.RSTICKPRESS, SCButtons.RSTICKPRESS)
			| map_button(self._input, DeckButton.LGRIP2, SCButtons.LGRIP2)
			| map_button(self._input, DeckButton.RGRIP2, SCButtons.RGRIP2)
		)
		# Convert triggers
		self._input.ltrig >>= 7
		self._input.rtrig >>= 7
		# Apply deadzones
		self._input.stick_x = apply_deadzone(self._input.stick_x, STICK_DEADZONE)
		self._input.stick_y = apply_deadzone(self._input.stick_y, STICK_DEADZONE)
		self._input.rstick_x = apply_deadzone(self._input.rstick_x, STICK_DEADZONE)
		self._input.rstick_y = apply_deadzone(self._input.rstick_y, STICK_DEADZONE)

		# Invert Gyro Roll to match Steam Controller coordinate system
		self._input.groll = -self._input.groll

		m = self.get_mapper()
		if m:
			self.mapper.input(self, self._old_state, self._input)

	def close(self):
		if self._ready:
			self.daemon.remove_controller(self)
			self._ready = False
		USBDevice.close(self)

	def turnoff(self):
		log.warning("Ignoring request to turn off steamdeck.")


def init(daemon, config):
	"""Register hotplug callback for controller dongle"""

	def cb(device, handle):
		return Deck(device, handle, daemon)

	register_hotplug_device(cb, VENDOR_ID, PRODUCT_ID)
	return True



import logging

from scc.actions import Action
from scc.parser import ActionParser, ParseError
from scc.tools import _

log = logging.getLogger("gui.parse")


class InvalidAction(Action):
	def __init__(self, string, error):
		self.string = string
		self.error = error
		self.name = None

	def __str__(self):
		return "<Invalid Action '%s'>" % (self.string,)

	__repr__ = __str__

	def to_string(self, *a):
		return self.string

	def describe(self, *a):
		return _("(invalid)")


class GuiActionParser(ActionParser):
	"""ActionParser that stores original string and
	returns InvalidAction instance when parsing fails
	"""

	def restart(self, string):
		self.string = string
		return ActionParser.restart(self, string)

	def parse(self):
		"""Returns parsed action or None if action cannot be parsed.
		"""
		try:
			a = ActionParser.parse(self)
			a.string = self.string
			return a
		except ParseError as e:
			log.error("Failed to parse '%s'", self.string)
			log.error(e)
			return InvalidAction(self.string, e)



"""SC Controller - Timer manager

Simple abstract class for named, cancelable timers
"""

from gi.repository import GLib


class TimerManager:
	def __init__(self) -> None:
		self._timers = {}

	def timer(self, name, delay, callback, *data, **kwdata) -> None:
		"""Run callback after specified number of seconds. Uses GLib.timeout_add_seconds with small wrapping to allow named timers to be canceled by reset() call"""
		method = GLib.timeout_add_seconds
		if delay < 1 and delay > 0:
			method = GLib.timeout_add
			delay = delay * 1000.0
		if name is None:
			# No wrapping is needed, call GLib directly
			method(delay, callback, *data, **kwdata)
		else:
			if name in self._timers:
				# Cancel old timer
				GLib.source_remove(self._timers[name])
			# Create new one
			self._timers[name] = method(delay, self._callback, name, callback, *data, **kwdata)

	def timer_active(self, name) -> bool:
		"""Return True if named timer is active"""
		return name in self._timers

	def cancel_timer(self, name) -> bool:
		"""Cancel named timer. Returns True on success, False if there is no such timer."""
		if name in self._timers:
			GLib.source_remove(self._timers[name])
			del self._timers[name]
			return True
		return False

	def cancel_all(self) -> None:
		"""Cancel all active timers"""
		for x in self._timers:
			GLib.source_remove(self._timers[x])
		self._timers = {}

	def _callback(self, name, callback, *data, **kwdata) -> bool:
		"""Remove name from list of active timers and calls real callback."""
		del self._timers[name]
		callback(*data, **kwdata)
		return False



#!/usr/bin/env python3
"""Imports VDFFZ profile and converts it to Profile object.
VDFFZ is just VDF encapsulated in json, so this just gets one value and calls
VDFProfile to decode rest.
"""

import json
import logging

from scc.lib.vdf import parse_vdf

from .vdf import VDFProfile

log = logging.getLogger("import.vdffz")


class VDFFZProfile(VDFProfile):
	def load(self, filename):
		try:
			data = json.loads(open(filename).read())
		except Exception:
			raise ValueError("Failed to parse JSON")
		if "ConfigData" not in data:
			raise ValueError("ConfigData missing in JSON")
		self.load_data(parse_vdf(data["ConfigData"].encode("utf-8")))



#!/usr/bin/env python3
"""SC-Controller - Action Editor - Per-Axis Component

Handles all XYActions
"""

import logging

from scc.actions import Action, NoAction, XYAction
from scc.gui.ae import AEComponent, describe_action
from scc.gui.area_to_action import action_to_area
from scc.gui.simple_chooser import SimpleChooser
from scc.tools import _

log = logging.getLogger("AE.PerAxis")

__all__ = ["PerAxisComponent"]


class PerAxisComponent(AEComponent):
	GLADE = "ae/per_axis.glade"
	NAME = "per_axis"
	CTXS = Action.AC_STICK | Action.AC_PAD
	PRIORITY = 1

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		self.x = self.y = NoAction()

	def set_action(self, mode, action):
		if isinstance(action, XYAction):
			self.x = action.x
			self.y = action.y
			self.update()

	def get_button_title(self):
		return _("Per Axis")

	def handles(self, mode, action):
		return isinstance(action, XYAction)

	def update(self):
		self.builder.get_object("lblAxisX").set_label(describe_action(Action.AC_STICK, None, self.x))
		self.builder.get_object("lblAxisY").set_label(describe_action(Action.AC_STICK, None, self.y))

	def send(self):
		self.editor.set_action(XYAction(self.x, self.y))

	def on_btAxisX_clicked(self, *a):
		"""'Select X Axis Action' handler"""

		def cb(action):
			self.x = action
			self.update()
			self.send()

		self.grab_action(self.x, cb)

	def on_btAxisY_clicked(self, *a):
		"""'Select Y Axis Action' handler"""

		def cb(action):
			self.y = action
			self.update()
			self.send()

		self.grab_action(self.y, cb)

	def grab_action(self, action, cb):
		b = SimpleChooser(self.app, "axis", cb)
		b.set_title(_("Select Axis"))
		area = action_to_area(action)
		b.display_action(Action.AC_STICK, area)
		b.show(self.editor.window)



"""SC Controller - Dualshock 4 Driver.

Extends HID driver with DS4-specific options.
"""
from __future__ import annotations

import ctypes
import logging
import os
import struct
import sys
import zlib
from typing import TYPE_CHECKING

import usb1
from evdev import ff

from scc.constants import STICK_PAD_MAX, STICK_PAD_MIN, ControllerFlags, HapticPos, SCButtons
from scc.controller import Controller
from scc.drivers.evdevdrv import (
	HAVE_EVDEV,
	EvdevController,
	get_axes,
	get_evdev_devices_from_syspath,
	make_new_device,
)
from scc.drivers.hiddrv import (
	BUTTON_COUNT,
	AxisData,
	AxisDataUnion,
	AxisMode,
	AxisModeData,
	AxisType,
	ButtonData,
	HatswitchModeData,
	HIDController,
	HIDDecoder,
	_lib,
	button_to_bit,
	hiddrv_test,
)
from scc.drivers.usb import USBDevice, register_hotplug_device
from scc.lib.hidraw import HIDRaw
from scc.tools import init_logging, set_logging_level

if TYPE_CHECKING:
	from evdev import InputDevice

	from scc.sccdaemon import SCCDaemon

log = logging.getLogger("DS4")

VENDOR_ID = 0x054C
PRODUCT_ID = 0x09CC
DS4_V1_PRODUCT_ID = 0x05C4

# CUH-ZCT2x exposes endpoint 0x03 as isochronous audio output and the HID
# interrupt output endpoint as 0x02.
DS4_USB_OUTPUT_ENDPOINT = 2
DS4_USB_OUTPUT_REPORT_SIZE = 32
DS4_USB_OUTPUT_REPORT_ID = 0x05
DS4_USB_OUTPUT_VALID_MOTOR = 0x01
DS4_BT_OUTPUT_REPORT_SIZE = 78
DS4_BT_OUTPUT_REPORT_ID = 0x11
DS4_BT_OUTPUT_HW_CONTROL = 0xC4
DS4_BT_OUTPUT_VALID_MOTOR = 0x01
DS4_BT_OUTPUT_CRC_SEED = 0xA2


class DS4Controller(Controller):
	# Most of axes are the same
	BUTTON_MAP = (
		SCButtons.X,
		SCButtons.A,
		SCButtons.B,
		SCButtons.Y,
		SCButtons.LB,
		SCButtons.RB,
		1 << 64,
		1 << 64,
		SCButtons.BACK,
		SCButtons.START,
		SCButtons.STICKPRESS,
		SCButtons.RPAD,
		SCButtons.C,
		SCButtons.CPADPRESS,
	)

	flags = (
		ControllerFlags.EUREL_GYROS
		| ControllerFlags.HAS_RSTICK
		| ControllerFlags.HAS_CPAD
		| ControllerFlags.HAS_DPAD
		| ControllerFlags.SEPARATE_STICK
		| ControllerFlags.NO_GRIPS
	)

	def __init__(self, daemon: SCCDaemon) -> None:
		self.daemon: SCCDaemon = daemon
		Controller.__init__(self)

	def _load_hid_descriptor(self, config, max_size, vid, pid, test_mode):
		# Overrided and hardcoded
		self._decoder = HIDDecoder()
		self._decoder.axes[AxisType.AXIS_LPAD_X] = AxisData(
			mode=AxisMode.HATSWITCH,
			byte_offset=5,
			size=8,
			data=AxisDataUnion(
				hatswitch=HatswitchModeData(
					button=SCButtons.LPAD | SCButtons.LPADTOUCH,
					min=STICK_PAD_MIN,
					max=STICK_PAD_MAX,
				),
			),
		)
		self._decoder.axes[AxisType.AXIS_STICK_X] = AxisData(
			mode=AxisMode.AXIS,
			byte_offset=1,
			size=8,
			data=AxisDataUnion(
				axis=AxisModeData(
					scale=1.0,
					offset=-127.5,
					clamp_max=257,
					deadzone=10,
				),
			),
		)
		self._decoder.axes[AxisType.AXIS_STICK_Y] = AxisData(
			mode=AxisMode.AXIS,
			byte_offset=2,
			size=8,
			data=AxisDataUnion(
				axis=AxisModeData(
					scale=-1.0,
					offset=127.5,
					clamp_max=257,
					deadzone=10,
				),
			),
		)
		self._decoder.axes[AxisType.AXIS_RPAD_X] = AxisData(
			mode=AxisMode.AXIS,
			byte_offset=3,
			size=8,
			data=AxisDataUnion(
				axis=AxisModeData(
					button=SCButtons.RPADTOUCH,
					scale=1.0,
					offset=-127.5,
					clamp_max=257,
					deadzone=10,
				),
			),
		)
		self._decoder.axes[AxisType.AXIS_RPAD_Y] = AxisData(
			mode=AxisMode.AXIS,
			byte_offset=4,
			size=8,
			data=AxisDataUnion(
				axis=AxisModeData(
					button=SCButtons.RPADTOUCH,
					scale=-1.0,
					offset=127.5,
					clamp_max=257,
					deadzone=10,
				),
			),
		)
		self._decoder.axes[AxisType.AXIS_LTRIG] = AxisData(
			mode=AxisMode.AXIS,
			byte_offset=8,
			size=8,
			data=AxisDataUnion(
				axis=AxisModeData(
					scale=1.0,
					clamp_max=1,
					deadzone=10,
				),
			),
		)
		self._decoder.axes[AxisType.AXIS_RTRIG] = AxisData(
			mode=AxisMode.AXIS,
			byte_offset=9,
			size=8,
			data=AxisDataUnion(
				axis=AxisModeData(
					scale=1.0,
					clamp_max=1,
					deadzone=10,
				),
			),
		)
		self._decoder.axes[AxisType.AXIS_GPITCH] = AxisData(mode=AxisMode.DS4ACCEL, byte_offset=13)
		self._decoder.axes[AxisType.AXIS_GROLL] = AxisData(mode=AxisMode.DS4ACCEL, byte_offset=17)
		self._decoder.axes[AxisType.AXIS_GYAW] = AxisData(mode=AxisMode.DS4ACCEL, byte_offset=15)
		self._decoder.axes[AxisType.AXIS_Q1] = AxisData(mode=AxisMode.DS4GYRO, byte_offset=23)
		self._decoder.axes[AxisType.AXIS_Q2] = AxisData(mode=AxisMode.DS4GYRO, byte_offset=19)
		self._decoder.axes[AxisType.AXIS_Q3] = AxisData(mode=AxisMode.DS4GYRO, byte_offset=21)

		self._decoder.axes[AxisType.AXIS_CPAD_X] = AxisData(mode=AxisMode.DS4TOUCHPAD, byte_offset=36)
		self._decoder.axes[AxisType.AXIS_CPAD_Y] = AxisData(mode=AxisMode.DS4TOUCHPAD, byte_offset=37, bit_offset=4)
		self._decoder.buttons = ButtonData(
			enabled=True,
			byte_offset=5,
			bit_offset=4,
			size=14,
			button_count=14,
		)

		if test_mode:
			for x in range(BUTTON_COUNT):
				self._decoder.buttons.button_map[x] = x
		else:
			for x in range(BUTTON_COUNT):
				self._decoder.buttons.button_map[x] = 64
			for x, sc in enumerate(DS4Controller.BUTTON_MAP):
				self._decoder.buttons.button_map[x] = button_to_bit(sc)

	def input(self, endpoint: int, data: bytearray) -> None:
		# Special override for CPAD touch button
		if _lib.decode(ctypes.byref(self._decoder), bytes(data)):
			if self.mapper:
				if data[35] >> 7:
					# cpad is not touched
					self._decoder.state.buttons &= ~SCButtons.CPADTOUCH
				else:
					self._decoder.state.buttons |= SCButtons.CPADTOUCH
				self.mapper.input(self, self._decoder.old_state, self._decoder.state)

	def get_gyro_enabled(self) -> bool:
		# Cannot be actually turned off, so it's always active
		# TODO: Maybe emulate turning off?
		return True

	def get_type(self) -> str:
		return "ds4"

	def get_gui_config_file(self) -> str:
		return "ds4-config.json"

	def __repr__(self) -> str:
		return f"<DS4Controller {self.get_id()}>"

	def _generate_id(self) -> str:
		"""ID is generated as 'ds4' or 'ds4:X' where 'X' starts as 1 and increases as controllers with same ids are connected."""
		magic_number = 1
		id = "ds4"
		while id in self.daemon.get_active_ids():
			id = f"ds4:{magic_number}"
			magic_number += 1
		return id


class DS4HIDController(DS4Controller, HIDController):
	def __init__(self, device: USBDevice, daemon: SCCDaemon, handle: usb1.USBDeviceHandle, config_file, config, test_mode=False) -> None:
		self._feedback_endpoint = self._find_feedback_endpoint(device)
		self._feedback_output = bytearray(DS4_USB_OUTPUT_REPORT_SIZE)
		self._feedback_output[0] = DS4_USB_OUTPUT_REPORT_ID
		self._feedback_output[1] = DS4_USB_OUTPUT_VALID_MOTOR
		self._feedback_pending = False
		self._feedback_cancel_tasks = [None, None]
		HIDController.__init__(self, device, daemon, handle, config_file, config, test_mode)
		log.debug("DS4 %s using USB/libusb", self.get_id())

	@staticmethod
	def _find_feedback_endpoint(device: USBDevice) -> int:
		"""Return the interrupt OUT endpoint belonging to the HID interface."""
		for interface in device[0]:
			for setting in interface:
				if setting.getClass() != 3:
					continue
				for endpoint in setting:
					address = endpoint.getAddress()
					if endpoint.getAttributes() == 3 and address & usb1.ENDPOINT_IN == 0:
						return address
		return DS4_USB_OUTPUT_ENDPOINT

	def feedback(self, data) -> None:
		position, amplitude, period, count = data.data
		amplitude = min(amplitude, 0x8000) / 0x8000
		right_amplitude = int(amplitude * 0xFF)
		left_amplitude = int(amplitude * 0xFF)

		motors = []
		if position in (HapticPos.LEFT, HapticPos.BOTH):
			self._feedback_output[5] = left_amplitude
			motors.append((0, 5))
		if position in (HapticPos.RIGHT, HapticPos.BOTH):
			self._feedback_output[4] = right_amplitude
			motors.append((1, 4))
		self._feedback_pending = True

		duration = max(float(period) * count / 0x10000, 0.02)
		for task_index, report_index in motors:
			task = self._feedback_cancel_tasks[task_index]
			if task:
				task.cancel()
			if amplitude == 0 or count == 0:
				self._feedback_cancel_tasks[task_index] = None
				continue

			def clear_feedback(mapper, task_index=task_index, report_index=report_index):
				self._feedback_output[report_index] = 0
				self._feedback_pending = True
				self._feedback_cancel_tasks[task_index] = None

			self._feedback_cancel_tasks[task_index] = self.mapper.schedule(duration, clear_feedback)

	def flush(self) -> None:
		HIDController.flush(self)
		if self._feedback_pending:
			self.handle.interruptWrite(self._feedback_endpoint, bytes(self._feedback_output))
			self._feedback_pending = False


class DS4HIDRawController(DS4Controller):
	def __init__(self, driver: DS4HIDRawDriver, syspath, hidrawdev: HIDRaw, vid, pid) -> None:
		self.driver = driver
		self.syspath = syspath

		DS4Controller.__init__(self, driver.daemon)

		self._device_name = hidrawdev.getName()
		self._hidrawdev = hidrawdev
		self._fileno = hidrawdev._device.fileno()
		self._id = self._generate_id() if driver else "-"
		self._closed: bool = False
		self._feedback_output = bytearray(DS4_BT_OUTPUT_REPORT_SIZE)
		self._feedback_output[0] = DS4_BT_OUTPUT_REPORT_ID
		self._feedback_output[1] = DS4_BT_OUTPUT_HW_CONTROL
		self._feedback_output[3] = DS4_BT_OUTPUT_VALID_MOTOR
		self._feedback_cancel_tasks = [None, None]

		self._packet_size = 78
		self._load_hid_descriptor(driver.config, self._packet_size, vid, pid, None)

		# self._set_operational()
		self.read_serial()
		self._poller = self.daemon.get_poller()
		if self._poller:
			self._poller.register(self._fileno, self._poller.POLLIN, self._input)
		self.daemon.get_device_monitor().add_remove_callback(syspath, self.close)
		self.daemon.add_controller(self)
		log.debug("DS4 %s using Bluetooth/hidraw (%s)", self.get_id(), self.syspath)

	def read_serial(self) -> None:
		self._serial = (self._hidrawdev.getPhysicalAddress().replace(b":", b""))

	def _input(self, *args) -> None:
		try:
			data = self._hidrawdev.read(self._packet_size)
		except Exception as error:
			log.debug("DS4 Bluetooth hidraw device disconnected: %s", error)
			self.close()
			return
		if data[0] != 0x11:
			return
		self.input(self._fileno, data[2:])

	def close(self, *args) -> None:
		if self._closed:
			return
		self._closed = True
		if self._poller:
			self._poller.unregister(self._fileno)

		self.daemon.remove_controller(self)
		self._hidrawdev._device.close()

	def _write_feedback_report(self) -> None:
		crc = zlib.crc32(bytes((DS4_BT_OUTPUT_CRC_SEED,)))
		crc = zlib.crc32(self._feedback_output[:-4], crc)
		struct.pack_into("<I", self._feedback_output, DS4_BT_OUTPUT_REPORT_SIZE - 4, crc)
		#log.debug("DS4 Bluetooth output: motors=(%s,%s)", self._feedback_output[7], self._feedback_output[6])
		self._hidrawdev.write(bytes(self._feedback_output))

	def feedback(self, data) -> None:
		position, amplitude, period, count = data.data
		amplitude = min(amplitude, 0x8000) / 0x8000
		motor_amplitude = int(amplitude * 0xFF)

		motors = []
		if position in (HapticPos.LEFT, HapticPos.BOTH):
			self._feedback_output[7] = motor_amplitude
			motors.append((0, 7))
		if position in (HapticPos.RIGHT, HapticPos.BOTH):
			self._feedback_output[6] = motor_amplitude
			motors.append((1, 6))

		duration = max(float(period) * count / 0x10000, 0.02)
		for task_index, report_index in motors:
			task = self._feedback_cancel_tasks[task_index]
			if task:
				task.cancel()
			if amplitude == 0 or count == 0:
				self._feedback_cancel_tasks[task_index] = None
				continue

			def clear_feedback(mapper, task_index=task_index, report_index=report_index):
				self._feedback_output[report_index] = 0
				self._feedback_cancel_tasks[task_index] = None
				self._write_feedback_report()

			self._feedback_cancel_tasks[task_index] = self.mapper.schedule(duration, clear_feedback)

		self._write_feedback_report()

	def turnoff(self) -> None:
		try:
			self.daemon.get_device_monitor().disconnect_bluetooth(self.syspath)
		except OSError as error:
			log.warning("Failed to turn off DS4 Bluetooth controller: %s", error)


class DS4HIDRawDriver:
	def __init__(self, daemon: SCCDaemon, config: dict) -> None:
		self.config = config
		self.daemon = daemon
		daemon.get_device_monitor().add_callback("bluetooth", VENDOR_ID, PRODUCT_ID, self.make_bt_hidraw_callback, None)
		daemon.get_device_monitor().add_callback("bluetooth", VENDOR_ID, DS4_V1_PRODUCT_ID, self.make_bt_hidraw_callback, None)

	def retry(self, syspath: str) -> None:
		pass

	def make_bt_hidraw_callback(self, syspath: str, vid, pid, *whatever) -> DS4HIDRawController | None:
		log.debug("DS4 Bluetooth callback: syspath=%s vid=%04x pid=%04x", syspath, vid, pid)
		hidrawname = self.daemon.get_device_monitor().get_hidraw(syspath)
		if hidrawname is None:
			return None
		try:
			dev = HIDRaw(open(os.path.join("/dev/", hidrawname), "w+b"))
			return DS4HIDRawController(self, syspath, dev, vid, pid)
		except Exception as e:
			log.exception(e)
			return None

	def get_device_name(self) -> str:
		return "Dualshock 4 over Bluetooth HIDRaw"

	def get_type(self) -> str:
		return "ds4bt_hidraw"

class DS4EvdevController(EvdevController):
	TOUCH_FACTOR_X = STICK_PAD_MAX / 940.0
	TOUCH_FACTOR_Y = STICK_PAD_MAX / 470.0
	BUTTON_MAP = {
		304: "A",
		305: "B",
		307: "Y",
		308: "X",
		310: "LB",
		311: "RB",
		314: "BACK",
		315: "START",
		316: "C",
		317: "STICKPRESS",
		318: "RPAD",
		# 319: "CPAD",
	}
	AXIS_MAP = {
		0: {"axis": "stick_x", "deadzone": 4, "max": 255, "min": 0},
		1: {"axis": "stick_y", "deadzone": 4, "max": 0, "min": 255},
		3: {"axis": "rpad_x", "deadzone": 4, "max": 255, "min": 0},
		4: {"axis": "rpad_y", "deadzone": 8, "max": 0, "min": 255},
		2: {"axis": "ltrig", "max": 255, "min": 0},
		5: {"axis": "rtrig", "max": 255, "min": 0},
		16: {"axis": "lpad_x", "deadzone": 0, "max": 1, "min": -1},
		17: {"axis": "lpad_y", "deadzone": 0, "max": -1, "min": 1},
	}
	BUTTON_MAP_OLD = {
		304: "X",
		305: "A",
		306: "B",
		307: "Y",
		308: "LB",
		309: "RB",
		312: "BACK",
		313: "START",
		314: "STICKPRESS",
		315: "RPAD",
		316: "C",
		# 317: "CPAD",
	}
	AXIS_MAP_OLD = {
		0: {"axis": "stick_x", "deadzone": 4, "max": 255, "min": 0},
		1: {"axis": "stick_y", "deadzone": 4, "max": 0, "min": 255},
		2: {"axis": "rpad_x", "deadzone": 4, "max": 255, "min": 0},
		5: {"axis": "rpad_y", "deadzone": 8, "max": 0, "min": 255},
		3: {"axis": "ltrig", "max": 32767, "min": -32767},
		4: {"axis": "rtrig", "max": 32767, "min": -32767},
		16: {"axis": "lpad_x", "deadzone": 0, "max": 1, "min": -1},
		17: {"axis": "lpad_y", "deadzone": 0, "max": -1, "min": 1},
	}
	GYRO_MAP = {
		EvdevController.ECODES.ABS_RX: ("gpitch", 0.01),
		EvdevController.ECODES.ABS_RY: ("gyaw", 0.01),
		EvdevController.ECODES.ABS_RZ: ("groll", 0.01),
		EvdevController.ECODES.ABS_X: (None, 1),  # 'q2'
		EvdevController.ECODES.ABS_Y: (None, 1),  # 'q3'
		EvdevController.ECODES.ABS_Z: (None, -1),  # 'q1'
	}
	flags = (
		ControllerFlags.EUREL_GYROS
		| ControllerFlags.HAS_RSTICK
		| ControllerFlags.HAS_CPAD
		| ControllerFlags.HAS_DPAD
		| ControllerFlags.SEPARATE_STICK
		| ControllerFlags.NO_GRIPS
	)

	def __init__(
		self,
		daemon: SCCDaemon,
		controllerdevice: InputDevice[str],
		gyro: InputDevice[str],
		touchpad: InputDevice[str],
		syspath: str,
	) -> None:
		config = {
			"axes": DS4EvdevController.AXIS_MAP,
			"buttons": DS4EvdevController.BUTTON_MAP,
			"dpads": {},
		}
		if controllerdevice.info.version & 0x8000 == 0:
			# Older kernel uses different mappings
			# see kernel source, drivers/hid/hid-sony.c#L2748
			config["axes"] = DS4EvdevController.AXIS_MAP_OLD
			config["buttons"] = DS4EvdevController.BUTTON_MAP_OLD
		self._gyro = gyro
		self._touchpad = touchpad
		self._syspath: str = syspath
		self._feedback_effect_id: int | None = None
		for device in (self._gyro, self._touchpad):
			if device:
				device.grab()
		EvdevController.__init__(self, daemon, controllerdevice, None, config)
		log.debug("DS4 %s using evdev (%s)", self.get_id(), controllerdevice.path)
		if self.poller:
			self.poller.register(touchpad.fd, self.poller.POLLIN, self._touchpad_input)
			self.poller.register(gyro.fd, self.poller.POLLIN, self._gyro_input)

	def _gyro_input(self, *a) -> None:
		new_state = self._state
		try:
			for event in self._gyro.read():
				if event.type == self.ECODES.EV_ABS:
					axis, factor = DS4EvdevController.GYRO_MAP[event.code]
					if axis:
						new_state = new_state._replace(**{axis: int(event.value * factor)})
		except OSError:
			# Errors here are not even reported, evdev class handles important ones
			return

		if new_state is not self._state:
			old_state, self._state = self._state, new_state
			if self.mapper:
				self.mapper.input(self, old_state, new_state)

	def _touchpad_input(self, *a):
		new_state = self._state
		try:
			for event in self._touchpad.read():
				if event.type == self.ECODES.EV_ABS:
					if event.code == self.ECODES.ABS_MT_POSITION_X:
						value = event.value * DS4EvdevController.TOUCH_FACTOR_X
						value = STICK_PAD_MIN + int(value)
						new_state = new_state._replace(cpad_x=value)
					elif event.code == self.ECODES.ABS_MT_POSITION_Y:
						value = event.value * DS4EvdevController.TOUCH_FACTOR_Y
						value = STICK_PAD_MAX - int(value)
						new_state = new_state._replace(cpad_y=value)
				elif event.type == 0:
					pass
				elif event.code == self.ECODES.BTN_LEFT:
					if event.value == 1:
						b = new_state.buttons | SCButtons.CPADPRESS
						new_state = new_state._replace(buttons=b)
					else:
						b = new_state.buttons & ~SCButtons.CPADPRESS
						new_state = new_state._replace(buttons=b)
				elif event.code == self.ECODES.BTN_TOUCH:
					if event.value == 1:
						b = new_state.buttons | SCButtons.CPADTOUCH
						new_state = new_state._replace(buttons=b)
					else:
						b = new_state.buttons & ~SCButtons.CPADTOUCH
						new_state = new_state._replace(buttons=b, cpad_x=0, cpad_y=0)
		except OSError:
			# Errors here are not even reported, evdev class handles important ones
			return

		if new_state is not self._state:
			old_state, self._state = self._state, new_state
			if self.mapper:
				self.mapper.input(self, old_state, new_state)

	def close(self) -> None:
		self._stop_feedback()
		EvdevController.close(self)
		for device in (self._gyro, self._touchpad):
			try:
				self.poller.unregister(device.fd)
				device.ungrab()
			except Exception:
				pass

	def get_gyro_enabled(self) -> bool:
		# Cannot be actually turned off, so it's always active
		# TODO: Maybe emulate turning off?
		return True

	def turnoff(self) -> None:
		if self.device.info.bustype != self.ECODES.BUS_BLUETOOTH:
			log.warning("Ignoring request to turn off wired DS4 controller")
			return
		try:
			self.daemon.get_device_monitor().disconnect_bluetooth(self._syspath)
		except Exception as error:
			log.warning("Failed to turn off DS4 evdev controller: %s", error)

	def _stop_feedback(self) -> None:
		if self._feedback_effect_id is None:
			return
		try:
			self.device.write(self.ECODES.EV_FF, self._feedback_effect_id, 0)
			self.device.erase_effect(self._feedback_effect_id)
		except OSError as error:
			log.debug("Failed to stop DS4 evdev rumble effect: %s", error)
		finally:
			self._feedback_effect_id = None

	def feedback(self, data) -> None:
		position, amplitude, period, count = data.data
		self._stop_feedback()
		if amplitude == 0 or count == 0:
			return

		magnitude = min(amplitude * 2, 0xFFFF)
		strong_magnitude = magnitude if position in (HapticPos.LEFT, HapticPos.BOTH) else 0
		weak_magnitude = magnitude if position in (HapticPos.RIGHT, HapticPos.BOTH) else 0
		duration = max(min(round(float(period) * count / 0x10000 * 1000), 0xFFFF), 20)
		effect = ff.Effect(
			self.ECODES.FF_RUMBLE,
			-1,
			0,
			ff.Trigger(0, 0),
			ff.Replay(duration, 0),
			ff.EffectType(
				ff_rumble_effect=ff.Rumble(
					strong_magnitude=strong_magnitude,
					weak_magnitude=weak_magnitude,
				),
			),
		)
		try:
			self._feedback_effect_id = self.device.upload_effect(effect)
			self.device.write(self.ECODES.EV_FF, self._feedback_effect_id, 1)
		except OSError as error:
			self._feedback_effect_id = None
			log.warning("Failed to play DS4 evdev rumble effect: %s", error)

	def get_type(self) -> str:
		return "ds4evdev"

	def get_gui_config_file(self) -> str:
		return "ds4-config.json"

	def __repr__(self) -> str:
		return f"<DS4EvdevController {self.get_id()}>"

	def _generate_id(self) -> str:
		"""ID is generated as 'ds4' or 'ds4:X' where 'X' starts as 1 and increases as controllers with same ids are connected."""
		magic_number = 1
		id = "ds4"
		while id in self.daemon.get_active_ids():
			id = f"ds4:{magic_number}"
			magic_number += 1
		return id


def init(daemon: SCCDaemon, config: dict) -> bool:
	"""Register hotplug callback for DS4 device."""

	def hid_callback(device, handle) -> DS4HIDController:
		return DS4HIDController(device, daemon, handle, None, None)

	def make_evdev_device(sys_dev_path: str, *whatever):
		devices = get_evdev_devices_from_syspath(sys_dev_path)
		# With kernel 4.10 or later, PS4 controller pretends to be 3 different devices.
		# 1st, determining which one is actual controller is needed
		controllerdevice = None
		for device in devices:
			count = len(get_axes(device))
			if count == 8:
				# 8 axes - Controller
				controllerdevice = device
		if not controllerdevice:
			log.debug("DS4 evdev controller node is not ready; discovery will retry")
			return None
		# 2nd, find motion sensor and touchpad with physical address matching controllerdevice
		gyro, touchpad = None, None
		phys = device.phys.split("/")[0]
		for device in devices:
			if device.phys.startswith(phys):
				axes = get_axes(device)
				count = len(axes)
				if count == 6:
					# 6 axes
					if EvdevController.ECODES.ABS_MT_POSITION_X in axes:
						# kernel 4.17+ - touchpad
						touchpad = device
					else:
						# gyro sensor
						gyro = device
				elif count == 4:
					# 4 axes - Touchpad
					touchpad = device
		# 3rd, do a magic
		if controllerdevice and gyro and touchpad:
			return make_new_device(DS4EvdevController, controllerdevice, gyro, touchpad, sys_dev_path)

	def fail_cb(syspath: str, vid: int, pid: int) -> None:
		if HAVE_EVDEV:
			log.warning("Failed to acquire USB device, falling back to evdev driver. This is far from optimal.")
			make_evdev_device(syspath)
		else:
			log.error(
				"Failed to acquire USB device and evdev is not available. Everything is lost and DS4 support disabled.",
			)
			# TODO: Maybe add_error here, but error reporting needs a little rework, so it's not treated as fatal
			# daemon.add_error("ds4", "No access to DS4 device")

	if config["drivers"].get("hiddrv") or (HAVE_EVDEV and config["drivers"].get("evdevdrv")):
		# DS4 v.2
		register_hotplug_device(hid_callback, VENDOR_ID, PRODUCT_ID, on_failure=fail_cb)
		# DS4 v.1
		register_hotplug_device(hid_callback, VENDOR_ID, DS4_V1_PRODUCT_ID, on_failure=fail_cb)
		if config["drivers"].get("hiddrv"):
			# Only enable HIDRaw support for BT connections if hiddrv is enabled
			_drv = DS4HIDRawDriver(daemon, config)
		elif HAVE_EVDEV and config["drivers"].get("evdevdrv"):
			# DS4 v.2
			daemon.get_device_monitor().add_callback("bluetooth", VENDOR_ID, PRODUCT_ID, make_evdev_device, None)
			# DS4 v.1
			daemon.get_device_monitor().add_callback("bluetooth", VENDOR_ID, DS4_V1_PRODUCT_ID, make_evdev_device, None)
		return True
	log.warning("Neither HID nor Evdev driver is enabled, DS4 support cannot be enabled.")
	return False


if __name__ == "__main__":
	""" Called when executed as script """
	init_logging()
	set_logging_level(True, True)
	sys.exit(hiddrv_test(DS4HIDController, ["054c:09cc"]))



"""SC Controller - Action Editor - "DPAD or Menu".

Setups DPAD emulation or menu display
"""

import logging

from scc.actions import (
	Action,
	ButtonAction,
	DPad8Action,
	DPadAction,
	HatDownAction,
	HatLeftAction,
	HatRightAction,
	HatUpAction,
	NoAction,
)
from scc.constants import DEFAULT, SAME, STICK, SCButtons
from scc.gui.ae import AEComponent, describe_action
from scc.gui.ae.menu_action import MenuActionCofC
from scc.gui.binding_editor import BindingEditor
from scc.tools import _
from scc.uinput import Axes, Keys

log = logging.getLogger("AE.DPAD")

__all__ = ["DPADComponent"]


class DPADComponent(AEComponent, MenuActionCofC, BindingEditor):
	GLADE = "ae/dpad.glade"
	NAME = "dpad"
	CTXS = Action.AC_STICK | Action.AC_PAD
	PRIORITY = 2

	DPAD8_WIDGETS = ["btDPAD4", "btDPAD5", "btDPAD6", "btDPAD7"]

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		MenuActionCofC.__init__(self)
		BindingEditor.__init__(self, app)
		self._recursing = False
		self._userdata_load_started = False
		self.actions = [NoAction()] * 8

	def load(self):
		if self.loaded:
			return
		AEComponent.load(self)
		cbConfirmWith = self.builder.get_object("cbConfirmWith")
		cbCancelWith = self.builder.get_object("cbCancelWith")
		cbConfirmWith.set_row_separator_func(lambda model, iter: model.get_value(iter, 0) == "-")
		cbCancelWith.set_row_separator_func(lambda model, iter: model.get_value(iter, 0) == "-")

	def shown(self):
		if not self._userdata_load_started:
			self._userdata_load_started = True
			self.load_menu_list()

	def set_action(self, mode, action):
		cbm = self.builder.get_object("cbMenuType")
		cb = self.builder.get_object("cbActionType")
		scl = self.builder.get_object("sclDiagonalRange")
		if isinstance(action, DPadAction):
			self.set_cb(cbm, "menu", 1)
			scl.set_value(action.diagonal_rage)
			if isinstance(action, DPad8Action):
				self.set_cb(cb, "dpad8", 1)
			else:
				self.set_cb(cb, "dpad", 1)
			self.update_button_desc(action)
		elif MenuActionCofC.handles(self, None, action):
			self.set_cb(cb, "menu", 1)
			self.load_menu_data(action)
		self.on_cbActionType_changed()

	def update_button_desc(self, action):
		for i in range(len(action.actions)):
			self.actions[i] = action.actions[i]
		for i in range(8):
			self.set_button_desc(i)

	def set_button_desc(self, i):
		desc = describe_action(Action.AC_BUTTON, None, self.actions[i])
		l = self.builder.get_object("lblDPAD%s" % (i,))
		if l is None:
			l = self.builder.get_object("btDPAD%s" % (i,)).get_children()[0]
		l.set_markup(desc)

	def get_button_title(self):
		return _("DPAD / Menu")

	def handles(self, mode, action):
		if MenuActionCofC.handles(self, mode, action):
			return True
		return isinstance(action, DPadAction)  # DPad8Action is derived from DPadAction

	def update(self):
		cb = self.builder.get_object("cbActionType")
		scl = self.builder.get_object("sclDiagonalRange")
		key = cb.get_model().get_value(cb.get_active_iter(), 1)
		if key == "dpad8":
			# 8-way dpad
			self.editor.set_action(DPad8Action(scl.get_value(), *self.actions))
		elif key == "dpad":
			# 4-way dpad
			self.editor.set_action(DPadAction(scl.get_value(), *self.actions[0:4]))
		elif key == "wsad":
			# special case of 4-way dpad
			a = DPadAction(
				scl.get_value(),
				ButtonAction(Keys.KEY_W),
				ButtonAction(Keys.KEY_S),
				ButtonAction(Keys.KEY_A),
				ButtonAction(Keys.KEY_D),
			)
			self.actions = [NoAction()] * 8
			self.editor.set_action(a)
			self.update_button_desc(a)
		elif key == "arrows":
			# special case of 4-way dpad
			a = DPadAction(
				scl.get_value(),
				ButtonAction(Keys.KEY_UP),
				ButtonAction(Keys.KEY_DOWN),
				ButtonAction(Keys.KEY_LEFT),
				ButtonAction(Keys.KEY_RIGHT),
			)
			self.actions = [NoAction()] * 8
			self.editor.set_action(a)
			self.update_button_desc(a)
		elif key == "actual_dpad":
			# maps to dpad as real gamepad usually has
			a = DPadAction(
				scl.get_value(),
				HatUpAction(Axes.ABS_HAT0Y),
				HatDownAction(Axes.ABS_HAT0Y),
				HatLeftAction(Axes.ABS_HAT0X),
				HatRightAction(Axes.ABS_HAT0X),
			)
			self.actions = [NoAction()] * 8
			self.editor.set_action(a)
			self.update_button_desc(a)
		else:
			# Menu
			self.on_cbMenus_changed()

	def on_cbActionType_changed(self, *a):
		if self._recursing:
			return
		cb = self.builder.get_object("cbActionType")
		stActionData = self.builder.get_object("stActionData")
		key = cb.get_model().get_value(cb.get_active_iter(), 1)
		if key in ("dpad", "dpad8", "wsad", "arrows", "actual_dpad"):
			for i in self.DPAD8_WIDGETS:
				self.builder.get_object(i).set_visible(key == "dpad8")
			stActionData.set_visible_child(self.builder.get_object("grDPAD"))
		else:  # key == "menu"
			stActionData.set_visible_child(self.builder.get_object("grMenu"))
		self.update()

	def on_action_chosen(self, i, action, mark_changed=True):
		self.actions[i] = action
		cb = self.builder.get_object("cbActionType")
		key = cb.get_model().get_value(cb.get_active_iter(), 1)
		if key != "dpad8":
			# When user chooses WSAD, Arrows or DPAD emulation and then changes
			# one of actions, swap back to 'Simple DPAD' mode.
			self.set_cb(cb, "dpad", 1)
		# if action.name:
		# action = NameModifier(action.name, action)
		self.set_button_desc(i)
		self.update()

	def on_sclDiagonalRange_format_value(self, scale, value):
		return _("%s°") % (value,)

	def on_btClearDiagonalRange_clicked(self, *a):
		scl = self.builder.get_object("sclDiagonalRange")
		scl.set_value(DPadAction.DEFAULT_DIAGONAL_RANGE)

	def on_sclDiagonalRange_value_changed(self, *a):
		self.update()

	def on_btDPAD_clicked(self, b):
		"""'Select DPAD Left Action' handler"""
		i = int(b.get_name())
		action = self.actions[i]
		# f isinstance(action, NameModifier):
		# action.action.name = action.name
		# action = action.action
		ae = self.choose_editor(action, "")
		# ae = ActionEditor(self.app, self.on_choosen)
		ae.set_title(_("Select DPAD Action"))
		ae.set_input(i, action, mode=Action.AC_BUTTON)
		ae.show(self.editor.window)

	def get_default_confirm(self):
		"""Returns default confirm button for pads/stick - LPAD, RPAD or STICKPRESS
		"""
		if self.editor.id == STICK:
			return SCButtons.STICKPRESS
		return getattr(SCButtons, self.editor.id)

	def get_default_cancel(self):
		"""Returns default cancel button for stick/pad - SAME or B
		"""
		if self.editor.id == STICK:
			return SCButtons.B
		return SAME

	def get_control_with(self):
		"""'control_with' argument is ignored when menu is used with stick/pad.
		"""
		return DEFAULT

	def on_exMenuControl_activate(self, ex, *a):
		rvMenuControl = self.builder.get_object("rvMenuControl")
		rvMenuControl.set_reveal_child(not ex.get_expanded())

	def on_exMenuPosition_activate(self, ex, *a):
		rvMenuPosition = self.builder.get_object("rvMenuPosition")
		rvMenuPosition.set_reveal_child(not ex.get_expanded())



#!/usr/bin/env python3
"""SC-Controller - Simple Chooser

Used by Action Editor to display window with just one Component
"""


import importlib
import logging

from scc.gui.ae import AEComponent
from scc.gui.dwsnc import headerbar
from scc.gui.editor import Editor

log = logging.getLogger("SimpleChooser")


class SimpleChooser(Editor):
	GLADE = "simple_chooser.glade"

	def __init__(self, app, component_name, callback):
		self.app = app
		self._action = None
		self.component = None
		self.callback = callback
		self.setup_widgets()
		self.load_component(component_name)

	def setup_widgets(self):
		Editor.setup_widgets(self)
		headerbar(self.builder.get_object("header"))

	def load_component(self, component_name):
		mod = importlib.import_module("scc.gui.ae.%s" % (component_name,))
		for x in dir(mod):
			cls = getattr(mod, x)
			if isinstance(cls, type) and issubclass(cls, AEComponent):
				if component_name == cls.NAME:
					self.component = cls(self.app, self)
					break
		if self.component is None:
			raise ValueError("Unknown component '%s'" % (component_name,))
		self.component.load()
		if component_name == "buttons":
			self.component.hide_toggle()
		self.window.add(self.component.get_widget())

	def display_action(self, mode, action):
		self._action = action
		self.component.set_action(mode, action)

	def set_action(self, action):
		self.callback(action)
		self.close()
		self.window.destroy()

	def hide_axes(self):
		"""Prevents user from selecting axes"""
		self.component.hide_axes()

	def hide_mouse(self):
		"""Prevents user from selecting mouse-related stuff"""
		self.component.hide_mouse()



"""SC-Controller - Action Editor - Gyro -> Joystick or Mouse component."""

import itertools
import logging
import re

from scc.actions import Action, GyroAbsAction, GyroAction, MouseAbsAction, MouseAction, MultiAction, NoAction, RangeOP
from scc.constants import ROLL, STICK, YAW, SCButtons
from scc.gui.ae import AEComponent
from scc.gui.parser import GuiActionParser
from scc.modifiers import ModeModifier, SensitivityModifier
from scc.special_actions import CemuHookAction
from scc.tools import _, nameof
from scc.uinput import Axes, Rels

log = logging.getLogger("AE.GyroAction")

__all__ = ["GyroActionComponent"]
TRIGGERS = (nameof(SCButtons.LT), nameof(SCButtons.RT))


class GyroActionComponent(AEComponent):
	GLADE = "ae/gyro_action.glade"
	NAME = "gyro_action"
	CTXS = Action.AC_GYRO
	PRIORITY = 3

	BUTTONS = (  # in order as displayed in combobox
		(None, _("Always Active")),
		(None, None),
		(SCButtons.LT, _("Left Trigger")),
		(SCButtons.RT, _("Right Trigger")),
		(SCButtons.LB, _("Left Bumper")),
		(SCButtons.RB, _("Right Bumper")),
		(None, None),
		(SCButtons.LPADTOUCH, _("Left Pad Touched")),
		(SCButtons.RPADTOUCH, _("Right Pad Touched")),
		(SCButtons.LPAD, _("Left Pad Pressed")),
		(SCButtons.RPAD, _("Right Pad Pressed")),
		(None, None),
		(SCButtons.LGRIP, _("Left Grip")),
		(SCButtons.RGRIP, _("Right Grip")),
		(STICK, _("Stick Tilted")),
		(None, None),
		(SCButtons.A, _("A")),
		(SCButtons.B, _("B")),
		(SCButtons.X, _("X")),
		(SCButtons.Y, _("Y")),
		(None, None),
		(SCButtons.BACK, _("Back (select)")),
		(SCButtons.C, _("Center")),
		(SCButtons.START, _("Start")),
		(None, None),
		(SCButtons.CPADTOUCH, _("Touch Touchpad")),
		(SCButtons.CPADPRESS, _("Press Touchpad")),
	)

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		self._recursing = False
		self.parser = GuiActionParser()

	def load(self) -> None:
		if self.loaded:
			return
		AEComponent.load(self)
		self._recursing = True
		cbGyroButton = self.builder.get_object("cbGyroButton")
		fill_buttons(cbGyroButton)
		self._recursing = False

	def set_action(self, mode: int, action) -> None:
		if self.handles(mode, action):
			if isinstance(action, NoAction):
				self.select_gyro_output("none")
				self.select_gyro_button(SCButtons.RPADTOUCH)
				return
			if isinstance(action, ModeModifier):
				self._recursing = True
				self.builder.get_object("cbInvertGyro").set_active(bool(action.default))
				self._recursing = False
				b = next(itertools.islice(action.mods.keys(), 0, 1))
				action = action.mods[b] or action.default
				self.select_gyro_button(b)
			else:
				self.select_gyro_button(None)
			if isinstance(action, SensitivityModifier) and isinstance(action.action, MouseAction):
				# Mouse (Desktop)
				self.select_gyro_output("mouse")
				if len(action.action.parameters) > 0 and action.action.parameters[0] == YAW:
					self.select_yaw_roll(YAW)
				else:
					self.select_yaw_roll(ROLL)
				self.editor.set_default_sensitivity(3.5, 3.5, 3.5)
				self.editor.set_sensitivity(*action.speeds)
			elif isinstance(action, MouseAction):
				# Mouse (Camera)
				self.select_gyro_output("mouse_cam")
				if len(action.parameters) > 0 and action.parameters[0] == YAW:
					self.select_yaw_roll(YAW)
				else:
					self.select_yaw_roll(ROLL)
			if isinstance(action, SensitivityModifier):
				self.editor.set_sensitivity(*action.speeds)
			elif isinstance(action, GyroAction):
				ap = action.parameters
				if len(ap) == 2:
					self.select_yaw_roll(YAW)
				else:
					self.select_yaw_roll(ROLL)
				if ap[0] == Axes.ABS_X and ap[-1] == Axes.ABS_Y:
					if isinstance(action, GyroAbsAction):
						self.select_gyro_output("left_abs")
					else:
						self.select_gyro_output("left")
				elif ap[0] == Axes.ABS_RX and ap[-1] == Axes.ABS_RY:
					if isinstance(action, GyroAbsAction):
						self.select_gyro_output("right_abs")
					else:
						self.select_gyro_output("right")
				elif ap[0] == Rels.REL_Y and ap[-1] == Rels.REL_X:
					self.select_gyro_output("mouse_stick")
			elif isinstance(action, CemuHookAction):
				self.select_gyro_output("cemuhook")
			self.modifier_updated()

	def modifier_updated(self) -> None:
		cbInvertY = self.builder.get_object("cbInvertY")
		sens = self.editor.get_sensitivity()
		inverted = len(sens) >= 2 and sens[1] < 0
		if cbInvertY.get_active() != inverted:
			self._recursing = True
			cbInvertY.set_active(inverted)
			self._recursing = False

		self.update()

	def cbInvertY_toggled_cb(self, cb, *a):
		if self._recursing:
			return
		sens = list(self.editor.get_sensitivity())
		# Ensure that editor accepts Y sensitivity
		if len(sens) >= 2:
			sens[1] = abs(sens[1])
			if cb.get_active():
				# Ensure that Y sensitivity is negative
				sens[1] *= -1
		self.editor.set_sensitivity(*sens)

	def get_button_title(self):
		return _("Joystick or Mouse")

	def handles(self, mode, action):
		if isinstance(action, NoAction):
			return True
		if is_gyro_enable(action):
			action = next(itertools.islice(action.mods.values(), 0, 1)) or action.default
			if isinstance(action, SensitivityModifier):
				action = action.action
		if isinstance(action, GyroAction):  # Takes GyroAbsAction as well
			ap = action.parameters
			if (len(ap) == 3 and not ap[1]) or len(ap) == 2:
				if (ap[0] == Axes.ABS_X and ap[-1] == Axes.ABS_Y) or (ap[0] == Axes.ABS_RX and ap[-1] == Axes.ABS_RY):
					return True
				if ap[0] == Rels.REL_Y and ap[-1] == Rels.REL_X:
					return True
			return False
		if isinstance(action, (MouseAction, MouseAbsAction, CemuHookAction)):
			return True
		return False

	def select_gyro_output(self, key):
		"""Just sets combobox value"""
		cb = self.builder.get_object("cbMode")
		model = cb.get_model()
		self._recursing = True
		for row in model:
			if key == row[2]:
				cb.set_active_iter(row.iter)
				self._recursing = False
				return
		self._recursing = False

	def select_yaw_roll(self, yawroll):
		"""Just sets combobox value"""
		cb = self.builder.get_object("cbYawRoll")
		model = cb.get_model()
		self._recursing = True
		for row in model:
			if yawroll == row[0]:
				cb.set_active_iter(row.iter)
				self._recursing = False
				return
		self._recursing = False

	def select_gyro_button(self, item):
		"""Just sets combobox value"""
		cb = self.builder.get_object("cbGyroButton")
		rvSoftLevel = self.builder.get_object("rvSoftLevel")
		sclSoftLevel = self.builder.get_object("sclSoftLevel")
		model = cb.get_model()
		self._recursing = True
		button = None
		if isinstance(item, RangeOP):
			button = nameof(item.what)
			sclSoftLevel.set_value(item.value)
			rvSoftLevel.set_reveal_child(True)
		elif item is not None:
			button = nameof(item.name)
		for row in model:
			if button == row[0] and row[1] != None:
				cb.set_active_iter(row.iter)
				self._recursing = False
				return
		self._recursing = False

	def on_cbInvertGyro_toggled(self, cb, *a):
		lblGyroEnable = self.builder.get_object("lblGyroEnable")
		if cb.get_active():
			lblGyroEnable.set_label(_("Gyro Disable Button"))
		else:
			lblGyroEnable.set_label(_("Gyro Enable Button"))
		if not self._recursing:
			self.send()

	def on_sclSoftLevel_format_value(self, scale, value):
		return "%s%%" % (int(value * 100.0),)

	def update(self, *a):
		cbMode = self.builder.get_object("cbMode")
		cbYawRoll = self.builder.get_object("cbYawRoll")
		lblYawRoll = self.builder.get_object("lblYawRoll")
		key = cbMode.get_model().get_value(cbMode.get_active_iter(), 2)
		cbYawRoll.set_sensitive(key != "cemuhook")
		lblYawRoll.set_sensitive(key != "cemuhook")

	def hidden(self):
		self.editor.set_default_sensitivity(1, 1, 1)

	def send(self, *a) -> None:
		if self._recursing:
			return

		cbMode = self.builder.get_object("cbMode")
		cbYawRoll = self.builder.get_object("cbYawRoll")
		rvSoftLevel = self.builder.get_object("rvSoftLevel")
		sclSoftLevel = self.builder.get_object("sclSoftLevel")
		cbGyroButton = self.builder.get_object("cbGyroButton")
		cbInvertGyro = self.builder.get_object("cbInvertGyro")
		cbInvertGyro = self.builder.get_object("cbInvertGyro")
		action = cbMode.get_model().get_value(cbMode.get_active_iter(), 0)
		key = cbMode.get_model().get_value(cbMode.get_active_iter(), 2)
		yawroll = cbYawRoll.get_model().get_value(cbYawRoll.get_active_iter(), 0)
		item = cbGyroButton.get_model().get_value(cbGyroButton.get_active_iter(), 0)
		rvSoftLevel.set_reveal_child(item in TRIGGERS)

		match = re.match(r"([^\[]+)\[([^\|]+)\|([^\]]+)\](.*)", action)
		if match:
			grps = match.groups()
			if yawroll == YAW:
				action = "%s%s%s" % (grps[0], grps[1], grps[3])
			else:
				action = "%s%s%s" % (grps[0], grps[2], grps[3])
		action = self.parser.restart(action).parse()

		if item and action:
			if item in TRIGGERS:
				what = RangeOP(getattr(SCButtons, item), ">=", sclSoftLevel.get_value())
			elif item == STICK:
				what = RangeOP(item, ">=", sclSoftLevel.get_value())
			else:
				what = getattr(SCButtons, item)
			if cbInvertGyro.get_active():
				action = ModeModifier(what, NoAction(), action)
			else:
				action = ModeModifier(what, action)
		if key == "mouse":
			self.editor.set_default_sensitivity(3.5, 3.5, 3.5)
		else:
			self.editor.set_default_sensitivity(1, 1, 1)

		self.update()
		self.editor.set_action(action)


def is_gyro_enable(modemod) -> bool:
	"""Return True if ModeModifier instance is used to create 'Gyro Enable Button'."""
	if isinstance(modemod, ModeModifier):
		if len(modemod.mods) != 1:
			return False
		action = list(modemod.mods.values())[0]
		if modemod.default:
			if not action:
				# Possibly, default action is gyro and mode is NoAction.
				# That would mean that Gyro Disable button mode is used.
				action = modemod.default
			else:
				return False
		if isinstance(action, SensitivityModifier):
			action = action.action
		if isinstance(action, ModeModifier):
			return False
		if isinstance(action, MouseAction):
			return True
		if isinstance(action, GyroAction):  # Takes GyroAbsAction as well
			return True
		if isinstance(action, MultiAction):
			for a in action.actions:
				if not isinstance(a, GyroAction):
					return False
			return True
	return False


def fill_buttons(cb):
	cb.set_row_separator_func(lambda model, iter: model.get_value(iter, 1) is None)
	model = cb.get_model()
	for button, text in GyroActionComponent.BUTTONS:
		model.append((None if button is None else nameof(button), text))
	cb.set_active(0)



#!/usr/bin/env python3
"""SC-Controller - Custom module loader

Loads ~/.config/scc/custom.py, if present. This allows injecting custom action
classes by user and breaking everything in very creative ways.

load_custom_module function needs to be called by daemon and GUI, so it exists
in separate module.
"""

import os

from scc.paths import get_config_path


def load_custom_module(log, who_calls="daemon"):
	"""Loads and imports ~/.config/scc/custom.py, if it is present and displays
	big, fat warning in such case.

	Returns True if file exists.
	"""
	filename = os.path.join(get_config_path(), "custom.py")
	if os.path.exists(filename):
		log.warning("=" * 60)
		log.warning("Loading %s" % (filename,))
		log.warning(
			"If you don't know what this means or you haven't created it, stop daemon right now and remove this file.",
		)
		log.warning("")
		log.warning("Also try removing it if %s crashes shortly after this message." % (who_calls,))

		import imp

		imp.load_source("custom", filename)
		log.warning("=" * 60)
		return True
	return False



"""SC Controller - Special Actions

Special Action is "special" since it cannot be handled by mapper alone.
Instead, on_sa_<actionname> method on handler instance set by
mapper.set_special_actions_handler() is called to do whatever action is supposed
to do. If handler is not set, or doesn't have reqiuired method defined,
action only prints warning to console.
"""
from __future__ import annotations

import logging
import sys
from difflib import get_close_matches
from math import sqrt
from typing import TYPE_CHECKING, Self

from scc.actions import (
	Action,
	HapticEnabledAction,
	OSDEnabledAction,
	SpecialAction,
)
from scc.constants import DEFAULT, LEFT, RIGHT, SAME, STICK, STICK_PAD_MAX, SCButtons
from scc.modifiers import Modifier
from scc.tools import clamp, nameof, strip_gesture

if TYPE_CHECKING:
	from collections.abc import Callable
	from typing import Any

log = logging.getLogger("SActions")
_: Callable[[str], str] = lambda x: x

class ChangeProfileAction(Action, SpecialAction):
	SA: str = "profile"
	COMMAND: str = SA

	def __init__(self, profile: str):
		Action.__init__(self, profile)
		self.profile: str = profile

	def describe(self, context: int):
		if self.name:
			return self.name
		if context == Action.AC_OSD:
			return _("Profile: %s") % (self.profile,)
		if context == Action.AC_SWITCHER:
			return _("Switch to %s") % (self.profile,)
		return _("Profile Change")

	def get_compatible_modifiers(self) -> int:
		return Action.MOD_OSD

	def to_string(self, multiline: bool = False, pad: int = 0) -> str:
		return (" " * pad) + "%s('%s')" % (self.COMMAND, self.profile)

	def button_release(self, mapper):
		# Execute only when button is released (executing this when button
		# is pressed would send following button_release event to another
		# action from loaded profile)
		self.execute(mapper)

	def whole(self, mapper, *a):
		self.execute(mapper)


class ShellCommandAction(Action, SpecialAction):
	SA: str = "shell"
	COMMAND: str = SA

	def __init__(self, command: str):
		# if type(command) == str:
		# command = command.decode("unicode_escape")
		# assert type(command) == unicode
		Action.__init__(self, command)
		self.command: str = command

	def describe(self, context: int) -> str:
		if self.name:
			return self.name
		return _("Execute Command")

	def get_compatible_modifiers(self) -> int:
		return Action.MOD_OSD

	def to_string(self, multiline: bool = False, pad: int = 0) -> str:
		return (" " * pad) + "%s('%s')" % (self.COMMAND, self.parameters[0])

	def button_press(self, mapper):
		# Executes only when button is pressed
		return self.execute(mapper)


class TurnOffAction(Action, SpecialAction):
	SA: str = "turnoff"
	COMMAND: str = SA

	def __init__(self):
		Action.__init__(self)

	def describe(self, context: int) -> str:
		if self.name:
			return self.name
		if context == Action.AC_OSD:
			return _("Turning controller OFF")
		return _("Turn Off the Controller")

	def to_string(self, multiline: bool = False, pad: int = 0) -> str:
		return (" " * pad) + "%s()" % (self.COMMAND,)

	def get_compatible_modifiers(self) -> int:
		return Action.MOD_OSD

	def button_release(self, mapper):
		# Execute only when button is released (executing this when button
		# is pressed would hold stuck any other action bound to same button,
		# as button_release is not sent after controller turns off)
		self.execute(mapper)

	def whole(self, mapper, *a):
		self.execute(mapper)


class RestartDaemonAction(Action, SpecialAction):
	SA: str = "restart"
	COMMAND: str = SA
	ALIASES: tuple[str] = ("exit",)

	def __init__(self):
		Action.__init__(self)

	def describe(self, context) -> str:
		if self.name:
			return self.name
		return _("Restart SCC-Daemon")

	def to_string(self, multiline: bool = False, pad: int = 0) -> str:
		return (" " * pad) + "%s()" % (self.COMMAND,)

	def button_release(self, mapper):
		# Execute only when button is released (for same reason as
		# TurnOffAction does)
		self.execute(mapper)


class LedAction(Action, SpecialAction):
	SA: str = "led"
	COMMAND: str = SA

	def __init__(self, brightness: float):
		Action.__init__(self, brightness)
		self.brightness: int = clamp(0, int(brightness), 100)

	def describe(self, context: int) -> str:
		if self.name:
			return self.name
		return _("Set LED brightness")

	def get_compatible_modifiers(self) -> int:
		return Action.MOD_OSD

	def button_press(self, mapper):
		# Execute only when button is pressed
		self.execute(mapper)


class OSDAction(Action, SpecialAction):
	"""Displays text in OSD, or, if used as modifier, displays action description
	and executes that action.
	"""

	SA: str = "osd"
	COMMAND: str = SA
	DEFAULT_TIMEOUT: int = 5
	DEFAULT_SIZE: int = 3
	PROFILE_KEY_PRIORITY: int = -5  # After XYAction, but before everything else

	def __init__(self, *parameters):
		Action.__init__(self, *parameters)
		self.action: Action | None = None
		self.timeout: float = self.DEFAULT_TIMEOUT
		self.size: int = self.DEFAULT_SIZE
		if len(parameters) > 1 and type(parameters[0]) in (int, float):
			# timeout parameter included
			self.timeout = float(parameters[0])
			parameters = parameters[1:]
		if len(parameters) > 1 and type(parameters[0]) in (int, float):
			# size parameter included
			self.size = int(parameters[0])
			parameters = parameters[1:]
		if isinstance(parameters[0], Action):
			self.action = parameters[0]
			self.text = self.action.describe(Action.AC_OSD)
		else:
			self.text = str(parameters[0])
		if self.action and isinstance(self.action, OSDEnabledAction):
			self.action.enable_osd(self.timeout)

	def get_compatible_modifiers(self) -> int:
		if self.action:
			return self.action.get_compatible_modifiers()
		return 0

	@staticmethod
	def decode(data: dict[str, Any], a, *b) -> OSDAction:
		a = OSDAction(a)
		if data["osd"] is not True:
			a.timeout = float(data["osd"])
		return a

	def describe(self, context: int) -> str:
		if self.name:
			return self.name
		if self.action:
			return _("%s (with OSD)") % (self.action.describe(context),)
		if context == Action.AC_OSD:
			return _("Display '%s'" % self.text)
		return _("OSD Message")

	def to_string(self, multiline: bool = False, pad: int = 0) -> str:
		parameters: list[str] = []
		if self.timeout != self.DEFAULT_TIMEOUT or self.size != self.DEFAULT_SIZE:
			parameters.append(str(self.timeout))
		if self.size != self.DEFAULT_SIZE:
			parameters.append(str(self.size))
		if self.action:
			parameters.append(self.action.to_string(multiline=multiline, pad=pad))
		else:
			parameters.append("'%s'" % (str(self.text),))
		return (" " * pad) + "%s(%s)" % (self.COMMAND, ",".join(parameters))

	def strip(self) -> Action | Self:
		if self.action:
			return self.action.strip()
		return self

	def compress(self):
		if self.action:
			if isinstance(self.action, OSDEnabledAction):
				return self.action.compress()
			self.action = self.action.compress()
		return self

	def button_press(self, mapper):
		self.execute(mapper)
		if self.action:
			return self.action.button_press(mapper)

	def button_release(self, mapper):
		if self.action:
			return self.action.button_release(mapper)

	def trigger(self, mapper, position, old_position):
		if self.action:
			return self.action.trigger(mapper, position, old_position)

	def axis(self, mapper, position, what):
		if self.action:
			return self.action.axis(mapper, position, what)

	def pad(self, mapper, position, what):
		if self.action:
			return self.action.pad(mapper, position, what)

	def whole(self, mapper, x, y, what):
		if self.action:
			return self.action.whole(mapper, x, y, what)


class ClearOSDAction(Action, SpecialAction):
	"""Clears all windows from OSD layer. Cancels all menus, clears all messages,
	etc, etc.
	"""

	SA: str = "clearosd"
	COMMAND: str = SA

	def describe(self, context: int) -> str:
		return _("Hide all OSD Menus and Messages")

	def button_press(self, mapper):
		self.execute(mapper)


class MenuAction(Action, SpecialAction, HapticEnabledAction):
	"""Displays menu defined in profile or globally.
	"""

	SA: str = "menu"
	COMMAND: str = SA
	MENU_TYPE: str = "menu"
	MIN_STICK_DISTANCE: float = STICK_PAD_MAX / 3
	DEFAULT_POSITION: tuple[int, int] = 10, -10

	def __init__(
		self,
		menu_id: str,
		control_with: str | int = DEFAULT,
		confirm_with: str = DEFAULT,
		cancel_with: str = DEFAULT,
		show_with_release: bool = False,
		size: int = 0,
	):
		if control_with == SAME:
			# Little touch of backwards compatibility
			control_with, confirm_with = DEFAULT, SAME
		if type(control_with) is int:
			# Allow short form in case when menu is assigned to pad
			# eg.: menu("some-id", 3) sets size to 3
			control_with, size = DEFAULT, control_with
		Action.__init__(self, menu_id, control_with, confirm_with, cancel_with, show_with_release, size)
		HapticEnabledAction.__init__(self)
		self.menu_id: str = menu_id
		self.control_with: str = control_with
		self.confirm_with: str = confirm_with
		self.cancel_with: str = cancel_with
		self.size: int = size
		self.x: float
		self.y: float
		self.x, self.y = MenuAction.DEFAULT_POSITION
		self.show_with_release: bool = bool(show_with_release)
		self._stick_distance: float = 0

	def describe(self, context) -> str:
		if self.name:
			return self.name
		return _("Menu")

	def get_compatible_modifiers(self) -> int:
		return Action.MOD_FEEDBACK

	def to_string(self, multiline: bool = False, pad: int = 0) -> str:
		if self.control_with == DEFAULT:
			dflt = (DEFAULT, DEFAULT, False)
			vals = (self.confirm_with, self.cancel_with, self.show_with_release)
			if dflt == vals:
				# Special case when menu is assigned to pad
				if self.size == 0:
					return "%s%s('%s')" % (" " * pad, self.COMMAND, self.menu_id)
				return "%s%s('%s', %s)" % (" " * pad, self.COMMAND, self.menu_id, self.size)

		return "%s%s(%s)" % (" " * pad, self.COMMAND, ",".join(Action.encode_parameters(self.strip_defaults())))

	def get_previewable(self) -> bool:
		return True

	def button_press(self, mapper):
		if not self.show_with_release:
			confirm_with = self.confirm_with
			cancel_with = self.cancel_with
			args = [mapper]
			if confirm_with == SAME:
				confirm_with = mapper.get_pressed_button() or DEFAULT
			elif confirm_with == DEFAULT:
				confirm_with = DEFAULT
			if cancel_with == DEFAULT:
				cancel_with = DEFAULT
			if nameof(self.control_with) in (LEFT, RIGHT):
				args += ["--use-cursor"]
			args += [
				"--control-with",
				nameof(self.control_with),
				"-x",
				str(self.x),
				"-y",
				str(self.y),
				"--size",
				str(self.size),
				"--confirm-with",
				nameof(confirm_with),
				"--cancel-with",
				nameof(cancel_with),
			]
			self.execute(*args)

	def button_release(self, mapper):
		if self.show_with_release:
			self.execute(mapper, "-x", str(self.x), "-y", str(self.y))

	def whole(self, mapper, x: float, y: float, what: str, *params):
		if x == 0 and y == 0:
			# Sent when pad is released - don't display menu then
			return
		if self.haptic:
			params = list(params) + ["--feedback-amplitude", str(self.haptic.get_amplitude())]
		if what in (LEFT, RIGHT):
			confirm_with = self.confirm_with
			cancel_with = self.cancel_with
			if what == LEFT:
				if confirm_with == DEFAULT:
					confirm_with = SCButtons.LPAD
				if cancel_with == DEFAULT:
					cancel_with = SCButtons.LPADTOUCH
			elif what == RIGHT:
				if confirm_with == DEFAULT:
					confirm_with = SCButtons.RPAD
				if cancel_with == DEFAULT:
					cancel_with = SCButtons.RPADTOUCH
			else:
				# Stick
				if confirm_with == DEFAULT:
					confirm_with = SCButtons.STICKPRESS
				if cancel_with == DEFAULT:
					cancel_with = SCButtons.B
			if not mapper.was_pressed(cancel_with):
				self.execute(
					mapper,
					"--control-with",
					what,
					"-x",
					str(self.x),
					"-y",
					str(self.y),
					"--use-cursor",
					"--size",
					str(self.size),
					"--confirm-with",
					nameof(confirm_with),
					"--cancel-with",
					nameof(cancel_with),
					*params,
				)
		if what == STICK:
			# Special case, menu is displayed only if is moved enought
			distance = sqrt(x * x + y * y)
			if self._stick_distance < MenuAction.MIN_STICK_DISTANCE and distance > MenuAction.MIN_STICK_DISTANCE:
				self.execute(
					mapper,
					"--control-with",
					STICK,
					"-x",
					str(self.x),
					"-y",
					str(self.y),
					"--use-cursor",
					"--size",
					str(self.size),
					"--confirm-with",
					"STICKPRESS",
					"--cancel-with",
					STICK,
					*params,
				)
			self._stick_distance = distance


class HorizontalMenuAction(MenuAction):
	"""Same as menu, but packed as row
	"""

	COMMAND: str = "hmenu"
	MENU_TYPE: str = "hmenu"

class GridMenuAction(MenuAction):
	"""Same as menu, but displayed in grid
	"""

	COMMAND: str = "gridmenu"
	MENU_TYPE: str = "gridmenu"

class QuickMenuAction(MenuAction):
	"""Quickmenu. Max.6 items, controller by buttons
	"""

	COMMAND: str = "quickmenu"
	MENU_TYPE: str = "quickmenu"

	def describe(self, context):
		if self.name:
			return self.name
		return _("QuickMenu")

	def button_press(self, mapper):
		# QuickMenu is always shown with release
		pass

	def button_release(self, mapper):
		self.execute(mapper, "-x", str(self.x), "-y", str(self.y))


class RadialMenuAction(MenuAction):
	"""Same as grid menu, which is same as menu but displayed in grid,
	but displayed as circle.
	"""

	COMMAND: str = "radialmenu"
	MENU_TYPE: str = "radialmenu"

	def __init__(
		self,
		menu_id: str,
		control_with: str | int = DEFAULT,
		confirm_with: str = DEFAULT,
		cancel_with: str = DEFAULT,
		show_with_release: bool = False,
		size: int = 0,
	):
		MenuAction.__init__(self, menu_id, control_with, confirm_with, cancel_with, show_with_release, size)
		self.rotation: float = 0

	def whole(self, mapper, x: float, y: float, what: str):
		if self.rotation:
			MenuAction.whole(self, mapper, x, y, what, "--rotation", self.rotation)
		else:
			MenuAction.whole(self, mapper, x, y, what)

	def set_rotation(self, angle: float):
		self.rotation = angle

	def get_compatible_modifiers(self) -> int:
		return MenuAction.get_compatible_modifiers(self) or Action.MOD_ROTATE


class DialogAction(Action, SpecialAction):
	"""Dialog is actually kind of menu, but options for it are different.
	"""

	SA: str = "dialog"
	COMMAND: str = SA
	DEFAULT_POSITION: tuple[int, int] = 10, -10

	def __init__(self, *pars):
		Action.__init__(self, pars)

		self.options = []
		self.confirm_with = DEFAULT
		self.cancel_with = DEFAULT
		self.text = _("Dialog")
		self.x, self.y = MenuAction.DEFAULT_POSITION
		# First and 2nd parameter may be confirm and cancel button
		if len(pars) > 0 and pars[0] in SCButtons.__members__.values():
			self.confirm_with, pars = pars[0], pars[1:]
			if len(pars) > 0 and pars[0] in SCButtons.__members__.values():
				self.cancel_with, pars = pars[0], pars[1:]
		# 1st always present argument is title
		if len(pars) > 0:
			self.text, pars = pars[0], pars[1:]
		# ... everything else are actions
		self.options = pars

	def describe(self, context) -> str:
		if self.name:
			return self.name
		return _("Dialog")

	def to_string(self, multiline: bool = False, pad: int = 0) -> str:
		rv = "%s%s(" % (" " * pad, self.COMMAND)
		if self.confirm_with != DEFAULT:
			rv += "%s, " % (nameof(self.confirm_with),)
			if self.cancel_with != DEFAULT:
				rv += "%s, " % (nameof(self.cancel_with),)
		rv += "'%s', " % (self.text,)
		if multiline:
			rv += "\n%s" % (" " * (pad + 2))
		for option in self.options:
			rv += "%s, " % (option.to_string(False),)
			if multiline:
				rv += "\n%s" % (" " * (pad + 2))

		rv = rv.strip("\n ,")
		if multiline:
			rv += "\n)"
		else:
			rv += ")"
		return rv

	def get_previewable(self) -> bool:
		return False

	def button_release(self, mapper):
		confirm_with = self.confirm_with
		cancel_with = self.cancel_with
		args = [
			mapper,
			"-x",
			str(self.x),
			"-y",
			str(self.y),
			"--confirm-with",
			nameof(confirm_with),
			"--cancel-with",
			nameof(cancel_with),
			"--text",
			self.text,
		]
		for x in self.options:
			args.append(x)
		self.execute(*args)


class KeyboardAction(Action, SpecialAction):
	"""Shows OSD keyboard.
	"""

	SA: str = "keyboard"
	COMMAND: str = SA

	def __init__(self):
		Action.__init__(self)

	def get_compatible_modifiers(self) -> int:
		return Action.MOD_POSITION

	def describe(self, context: int) -> str:
		if self.name:
			return self.name
		if context == Action.AC_OSD:
			return _("Display Keyboard")
		return _("OSD Keyboard")

	def to_string(self, multiline: bool = False, pad: int = 0) -> str:
		return (" " * pad) + "%s()" % (self.COMMAND,)

	def button_press(self, mapper):
		self.execute(mapper)


class PositionModifier(Modifier):
	"""Sets position for OSD menu.
	"""

	COMMAND: str = "position"

	def _mod_init(self, x: int, y: int):
		self.position: tuple[int, int] = (x, y)

	def compress(self):
		if isinstance(self.action, MenuAction):
			self.action.x, self.action.y = self.position
		return self.action

	@staticmethod
	def decode(data: dict[str, Any], a, *b) -> PositionModifier:
		x, y = data[PositionModifier.COMMAND]
		return PositionModifier(x, y, a)

	def describe(self, context: int) -> str:
		return self.action.describe(context)


class GesturesAction(Action, OSDEnabledAction, SpecialAction):
	"""Stars gesture detection on pad. Recognition is handled by whatever
	is special_actions_handler and results are then sent back to this action
	as parameter of gesture() method.
	"""

	SA: str = "gestures"
	COMMAND: str = SA
	PROFILE_KEYS: tuple[str] = ("gestures",)
	PROFILE_KEY_PRIORITY: int = 2
	DEFAULT_PRECISION: float = 1.0

	def __init__(self, *stuff):
		OSDEnabledAction.__init__(self)
		Action.__init__(self, *stuff)
		self.gestures = {}
		self.precision = self.DEFAULT_PRECISION
		gstr = None

		if len(stuff) > 0 and type(stuff[0]) in (int, float):
			self.precision = clamp(0.0, float(stuff[0]), 1.0)
			stuff = stuff[1:]

		for i in stuff:
			if gstr is None and type(i) == str:
				gstr = i
			elif gstr is not None and isinstance(i, Action):
				self.gestures[gstr] = i
				gstr = None
			else:
				raise ValueError("Invalid parameter for '%s': unexpected %s" % (self.COMMAND, i))

	def get_compatible_modifiers(self) -> int:
		return Action.MOD_OSD

	def describe(self, context: int) -> str:
		if self.name:
			return self.name
		return _("Gestures")

	def to_string(self, multiline=False, pad=0):
		if multiline:
			rv = [(" " * pad) + self.COMMAND + "("]
			if self.precision != self.DEFAULT_PRECISION:
				rv[0] += "%s," % (self.precision)
			for gstr in self.gestures:
				a_str: list[str] = self.gestures[gstr].to_string(True).split("\n")
				a_str[0] = (" " * pad) + "  '" + (gstr + "',").ljust(11) + a_str[0]  # Key has to be one of SCButtons
				for i in range(1, len(a_str)):
					a_str[i] = (" " * pad) + "  " + a_str[i]
				a_str[-1] = a_str[-1] + ","
				rv += a_str
			if rv[-1][-1] == ",":
				rv[-1] = rv[-1][0:-1]
			rv += [(" " * pad) + ")"]
			return "\n".join(rv)
		rv = []
		if self.precision != self.DEFAULT_PRECISION:
			rv.append(str(self.precision))
		for gstr in self.gestures:
			rv += ["'%s'" % (gstr,), self.gestures[gstr].to_string(False)]
		return self.COMMAND + "(" + ", ".join(rv) + ")"

	def compress(self) -> Self:
		for gstr in self.gestures:
			a = self.gestures[gstr].compress()
			if "i" in gstr:
				del self.gestures[gstr]
				gstr = strip_gesture(gstr)
			self.gestures[gstr] = a
		return self

	@staticmethod
	def decode(data, a, parser, *b) -> OSDAction | GesturesAction:
		ga = GesturesAction()
		ga.gestures = {
			gstr: parser.from_json_data(data[GesturesAction.PROFILE_KEYS[0]][gstr])
			for gstr in data[GesturesAction.PROFILE_KEYS[0]]
		}
		if "name" in data:
			ga.name = data["name"]
		if "osd" in data:
			ga = OSDAction(ga)
		return ga

	def _find_exact_gesture(self, gesture_string: str):
		return self.gestures.get(gesture_string)

	def _find_ignore_stroke_count_gesture(self, gesture_string: str):
		stripped_gesture_string = strip_gesture(gesture_string)
		return self.gestures.get(stripped_gesture_string)

	def _find_best_match_gesture(self, gesture_string: str):
		NUM_MATCHES_TO_RETURN = 1

		similar_gestures = get_close_matches(
			gesture_string, self.gestures.keys(), NUM_MATCHES_TO_RETURN, self.precision,
		)
		best_gesture = next(iter(similar_gestures), None)

		if best_gesture is not None:
			return self.gestures[best_gesture]
		return None

	def find_gesture_action(self, gesture_string: str):
		action = None
		action = action or self._find_exact_gesture(gesture_string)
		action = action or self._find_ignore_stroke_count_gesture(gesture_string)
		action = action or self._find_best_match_gesture(gesture_string)
		return action

	def gesture(self, mapper, gesture_string: str):
		action = self.find_gesture_action(gesture_string)
		if action:
			action.button_press(mapper)
			mapper.schedule(0, action.button_release)

	def whole(self, mapper, x, y, what):
		if (x, y) != (0, 0):
			# (0, 0) singlanizes released touchpad
			self.execute(mapper, x, y, what)


class CemuHookAction(Action, SpecialAction):
	SA: str = "cemuhook"
	COMMAND: str = SA
	MAGIC_GYRO: float = 2000.0 / 32768.0
	ACC_RES_PER_G: float = 16384.0

	def gyro(self, mapper, *pyr):
		sa_data = (
			-mapper.state.accel_x / CemuHookAction.ACC_RES_PER_G,  # AccelX
			-mapper.state.accel_z / CemuHookAction.ACC_RES_PER_G,  # AccelZ
			mapper.state.accel_y / CemuHookAction.ACC_RES_PER_G,  # AccelY
			pyr[0] * CemuHookAction.MAGIC_GYRO,  # Gyro Pitch
			-pyr[1] * CemuHookAction.MAGIC_GYRO,  # Gyro Yaw
			-pyr[2] * CemuHookAction.MAGIC_GYRO,  # Gyro Roll
		)
		# log.debug(sa_data)
		self.execute(mapper, sa_data)

	def describe(self, context: int) -> str:
		if self.name:
			return self.name
		return _("CemuHook")


# Register actions from current module
Action.register_all(sys.modules[__name__])
